"""
Tests of IBD finding algorithms.
"""
import collections
import io
import itertools

import msprime
import numpy as np
import pytest

import tests
import tests.ibd as ibd
import tests.test_wright_fisher as wf
import tskit
from tests.test_highlevel import get_example_tree_sequences


# ↑ See https://github.com/tskit-dev/tskit/issues/1804 for when
# we can remove this. The example_ts here is intended to be the
# basic tree sequence which should give a meaningful result for
# most operations. Probably rename it to ``examples.simple_ts()``
# or something.


@tests.cached_example
def example_ts():
    return [msprime.sim_ancestry(2, random_seed=1)]


def ibd_segments(
    ts,
    *,
    within=None,
    between=None,
    min_length=0,
    max_time=None,
    compare_lib=True,
    print_c=False,
    print_py=False,
):
    """
    Calculates IBD segments using Python and converts output to lists of segments.
    Also compares result with C library.
    """
    ibd_f = ibd.IbdFinder(
        ts, within=within, between=between, max_time=max_time, min_length=min_length
    )
    ibd_segs = ibd_f.run()
    if print_py:
        print("Python output:\n")
        print(ibd_segs)
    # ibd_f.print_state()
    if compare_lib:
        c_out = ts.ibd_segments(
            within=within,
            between=between,
            max_time=max_time,
            min_length=min_length,
            store_segments=True,
        )
        if print_c:
            print("C output:\n")
            print(c_out)
        assert_ibd_equal(ibd_segs, c_out)
    return ibd_segs


def naive_ibd(ts, a, b):
    """
    Returns the IBD segments along the genome for a and b.
    """

    def path(tree, u, v):
        ret = [u]
        while u != v:
            u = tree.parent(u)
            ret.append(u)
        return ret

    tree = ts.first()
    mrca = tree.mrca(a, b)
    last_paths = [path(tree, a, mrca), path(tree, b, mrca)]
    last_mrca = mrca
    left = 0.0
    segs = []
    while tree.next():
        mrca = tree.mrca(a, b)
        paths = [path(tree, a, mrca), path(tree, b, mrca)]
        if paths != last_paths:
            segs.append(tskit.IdentitySegment(left, tree.interval.left, last_mrca))
            last_paths = paths
            left = tree.interval.left
            last_mrca = mrca

    segs.append(tskit.IdentitySegment(left, ts.sequence_length, last_mrca))
    # Filter out segments with no mrca
    return [seg for seg in segs if seg.node != -1]


def naive_ibd_all_pairs(ts, samples=None):
    samples = ts.samples() if samples is None else samples
    all_pairs_map = {
        (a, b): naive_ibd(ts, a, b) for a, b in itertools.combinations(samples, 2)
    }
    # Filter out pairs with empty segment lists
    return {key: value for key, value in all_pairs_map.items() if len(value) > 0}


class TestIbdDefinition:
    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_all_pairs(self, ts):
        samples = ts.samples()[:10]
        ibd_lib = ts.ibd_segments(within=samples, store_segments=True)
        ibd_def = naive_ibd_all_pairs(ts, samples=samples)
        assert_ibd_equal(ibd_lib, ibd_def)

    @pytest.mark.parametrize("N", [2, 5, 10])
    @pytest.mark.parametrize("T", [2, 5, 10])
    def test_wright_fisher_examples(self, N, T):
        tables = wf.wf_sim(N, T, deep_history=False, seed=42)
        tables.sort()
        # NB this is essential! We get spurious breakpoints otherwise
        tables.edges.squash()
        tables.sort()
        ts = tables.tree_sequence()
        ibd0 = ibd_segments(ts)
        ibd1 = naive_ibd_all_pairs(ts)
        assert_ibd_equal(ibd0, ibd1)


class TestIbdImplementations:
    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_all_pairs(self, ts):
        # Automatically compares the two implementations
        ibd_segments(ts)


def assert_ibd_equal(dict1, dict2):
    """
    Verifies that two dictionaries have the same keys, and that
    the set of items corresponding to each key is identical.
    Used to check identical IBD output.
    """
    assert len(dict1) == len(dict2)
    for key, val in dict1.items():
        assert key in dict2
        assert len(val) == len(dict2[key])
        segs1 = list(sorted(val))
        segs2 = list(sorted(dict2[key]))
        assert segs1 == segs2


class TestIbdSingleBinaryTree:
    @tests.cached_example
    def ts(self):
        #
        # 2        4
        #         / \
        # 1      3   \
        #       / \   \
        # 0    0   1   2
        print("evaluating ts")
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       1           0
        3       0           1
        4       0           2
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       1       3       0,1
        0       1       4       2,3
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    # Basic test
    def test_defaults(self):
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)],
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
            (1, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
        }
        ibd_segs = ibd_segments(self.ts(), within=[0, 1, 2])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_within(self):
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)],
        }
        ibd_segs = ibd_segments(self.ts(), within=[0, 1])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_between_0_1(self):
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)],
        }
        ibd_segs = ibd_segments(self.ts(), between=[[0], [1]])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_between_0_2(self):
        true_segs = {
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
        }
        ibd_segs = ibd_segments(self.ts(), between=[[0], [2]])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_between_0_1_2(self):
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)],
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
            (1, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
        }
        ibd_segs = ibd_segments(self.ts(), between=[[0], [1], [2]])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_between_0_12(self):
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)],
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 4)],
        }
        ibd_segs = ibd_segments(self.ts(), between=[[0], [1, 2]])
        assert_ibd_equal(ibd_segs, true_segs)

    def test_time(self):
        ibd_segs = ibd_segments(
            self.ts(),
            max_time=1.5,
            compare_lib=True,
        )
        true_segs = {(0, 1): [tskit.IdentitySegment(0.0, 1.0, 3)]}
        assert_ibd_equal(ibd_segs, true_segs)

    def test_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=2)
        assert_ibd_equal(ibd_segs, {})


class TestIbdInterface:
    @pytest.mark.parametrize("ts", example_ts())
    def test_input_errors_within(self, ts):
        with pytest.raises(tskit.LibraryError, match="Node out of bounds"):
            ts.ibd_segments(within=[-1])
        with pytest.raises(tskit.LibraryError, match="Duplicate sample value"):
            ts.ibd_segments(within=[0, 0])

    @pytest.mark.parametrize("ts", example_ts())
    def test_input_errors_between(self, ts):
        with pytest.raises(tskit.LibraryError, match="Node out of bounds"):
            ts.ibd_segments(between=[[0], [-1]])
        with pytest.raises(tskit.LibraryError, match="Duplicate sample"):
            ts.ibd_segments(between=[[0], [0]])

    @pytest.mark.parametrize("ts", example_ts())
    def test_within_between_mutually_exclusive(self, ts):
        with pytest.raises(ValueError, match="mutually exclusive"):
            ts.ibd_segments(within=[0], between=[1])

    @pytest.mark.parametrize("ts", example_ts())
    def test_tables_interface(self, ts):
        ibd_tab = ts.tables.ibd_segments(store_segments=True)
        ibd_ts = ts.ibd_segments(store_segments=True)
        assert ibd_tab == ibd_ts

    @pytest.mark.parametrize("ts", example_ts())
    def test_empty_within(self, ts):
        ibd = ts.ibd_segments(within=[], store_pairs=True)
        assert len(ibd) == 0

    @pytest.mark.parametrize("ts", example_ts())
    def test_empty_between(self, ts):
        ibd = ts.ibd_segments(between=[], store_pairs=True)
        assert len(ibd) == 0

    @pytest.mark.parametrize("ts", example_ts())
    def test_empty_in_between(self, ts):
        ibd = ts.ibd_segments(between=[[1, 2], []], store_pairs=True)
        assert len(ibd) == 0


class TestIbdTwoSamplesTwoTrees:

    # 2
    #             |     3
    # 1      2    |    / \
    #       / \   |   /   \
    # 0    0   1  |  0     1
    # |------------|----------|
    # 0.0          0.4        1.0

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       0           1
        3       0           1.5
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       0.4     2       0,1
        0.4     1.0     3       0,1
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    # Basic test
    def test_basic(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 1): [
                tskit.IdentitySegment(0.0, 0.4, 2),
                tskit.IdentitySegment(0.4, 1.0, 3),
            ]
        }
        assert_ibd_equal(ibd_segs, true_segs)

    # Max time = 1.2
    def test_time(self):
        ibd_segs = ibd_segments(self.ts(), max_time=1.2, compare_lib=True)
        true_segs = {(0, 1): [tskit.IdentitySegment(0.0, 0.4, 2)]}
        assert_ibd_equal(ibd_segs, true_segs)

    # Min length = 0.5
    def test_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=0.5, compare_lib=True)
        true_segs = {(0, 1): [tskit.IdentitySegment(0.4, 1.0, 3)]}
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdUnrelatedSamples:

    #
    #    2   3
    #    |   |
    #    0   1

    @tests.cached_example
    def ts(self):

        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       0           1
        3       0           1
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       1       2       0
        0       1       3       1
        """
        )

        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_basic(self):
        ibd_segs = ibd_segments(self.ts())
        assert len(ibd_segs) == 0

    def test_time(self):
        ibd_segs = ibd_segments(self.ts(), max_time=1.2)
        assert len(ibd_segs) == 0

    def test_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=0.2)
        assert len(ibd_segs) == 0


class TestIbdNoSamples:
    #
    #     2
    #    / \
    #   /   \
    #  /     \
    # (0)   (1)

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       0           0
        1       0           0
        2       0           1
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       1       2       0
        0       1       2       1
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_defaults(self):
        result = ibd_segments(self.ts())
        assert len(result) == 0

    def test_specified_samples(self):
        ibd_segs = ibd_segments(self.ts(), within=[0, 1])
        true_segs = {
            (0, 1): [
                tskit.IdentitySegment(0.0, 1, 2),
            ]
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdSamplesAreDescendants:
    #
    # 4     5
    # |     |
    # 2     3
    # |     |
    # 0     1
    #
    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       1           1
        3       1           1
        4       0           2
        5       0           2
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       1       2       0
        0       1       3       1
        0       1       4       2
        0       1       5       3
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_basic(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 2)],
            (1, 3): [tskit.IdentitySegment(0.0, 1.0, 3)],
        }

        assert_ibd_equal(ibd_segs, true_segs)

    def test_input_within(self):
        ibd_segs = ibd_segments(self.ts(), within=[0, 2, 3, 5])
        true_segs = {
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 2)],
            (3, 5): [tskit.IdentitySegment(0.0, 1.0, 5)],
        }
        assert_ibd_equal(ibd_segs, true_segs)

    def test_all_samples(self):
        # FIXME
        ibd_segs = ibd_segments(self.ts(), within=range(6), compare_lib=False)
        true_segs = {
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 2)],
            (0, 4): [tskit.IdentitySegment(0.0, 1.0, 4)],
            (2, 4): [tskit.IdentitySegment(0.0, 1.0, 4)],
            (1, 3): [tskit.IdentitySegment(0.0, 1.0, 3)],
            (1, 5): [tskit.IdentitySegment(0.0, 1.0, 5)],
            (3, 5): [tskit.IdentitySegment(0.0, 1.0, 5)],
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdSimpleInternalSampleChain:
    #
    # 2
    # |
    # 1
    # |
    # 0
    #
    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           1
        2       1           2
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       1       1       0
        0       1       2       1
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_basic(self):
        # FIXME
        ibd_segs = ibd_segments(self.ts(), compare_lib=False)
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0.0, 1.0, 1)],
            (0, 2): [tskit.IdentitySegment(0.0, 1.0, 2)],
            (1, 2): [tskit.IdentitySegment(0.0, 1.0, 2)],
        }

        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdDifferentPaths:
    #
    #        4       |      4       |        4
    #       / \      |     / \      |       / \
    #      /   \     |    /   3     |      /   \
    #     /     \    |   2     \    |     /     \
    #    /       \   |  /       \   |    /       \
    #   0         1  | 0         1  |   0         1
    #                |              |
    #                0.2            0.7

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       0           1
        3       0           1.5
        4       0           2.5
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0.2     0.7     2       0
        0.2     0.7     3       1
        0.0     0.2     4       0
        0.0     0.2     4       1
        0.7     1.0     4       0
        0.7     1.0     4       1
        0.2     0.7     4       2
        0.2     0.7     4       3
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_defaults(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 1): [
                tskit.IdentitySegment(0.0, 0.2, 4),
                tskit.IdentitySegment(0.7, 1.0, 4),
                tskit.IdentitySegment(0.2, 0.7, 4),
            ]
        }
        assert_ibd_equal(ibd_segs, true_segs)

    def test_time(self):
        ibd_segs = ibd_segments(self.ts(), max_time=1.8)
        assert len(ibd_segs) == 0

    def test_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=0.4)
        true_segs = {(0, 1): [tskit.IdentitySegment(0.2, 0.7, 4)]}
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdDifferentPaths2:
    #
    #        5         |
    #       / \        |
    #      /   4       |      4
    #     /   / \      |     / \
    #    /   /   \     |    /   \
    #   /   /     \    |   3     \
    #  /   /       \   |  / \     \
    # 0   1         2  | 0   2     1
    #                  |
    #                  0.2

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       1           0
        3       0           1
        4       0           2.5
        5       0           3.5
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0.2     1.0     3       0
        0.2     1.0     3       2
        0.0     1.0     4       1
        0.0     0.2     4       2
        0.2     1.0     4       3
        0.0     0.2     5       0
        0.0     0.2     5       4
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_defaults(self):
        ibd_segs = ibd_segments(self.ts(), within=[1, 2])
        true_segs = {
            (1, 2): [
                tskit.IdentitySegment(0.0, 0.2, 4),
                tskit.IdentitySegment(0.2, 1.0, 4),
            ],
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdDifferentPaths3:
    # 2.00┊   4   ┊   4   ┊
    #     ┊ ┏━╋━┓ ┊ ┏━╋━┓ ┊
    # 1.00┊ 2 ┃ 3 ┊ 3 ┃ 2 ┊
    #     ┊ ┃ ┃   ┊ ┃ ┃   ┊
    # 0.00┊ 0 1   ┊ 0 1   ┊
    #     0       5      10
    @tests.cached_example
    def ts(self):
        t = tskit.TableCollection(sequence_length=10)
        t.nodes.add_row(flags=1, time=0)
        t.nodes.add_row(flags=1, time=0)
        t.nodes.add_row(flags=0, time=1)
        t.nodes.add_row(flags=0, time=1)
        t.nodes.add_row(flags=0, time=2)
        t.edges.add_row(parent=2, child=0, left=0, right=5)
        t.edges.add_row(parent=3, child=0, left=5, right=10)
        t.edges.add_row(parent=4, child=2, left=0, right=10)
        t.edges.add_row(parent=4, child=3, left=0, right=10)
        t.edges.add_row(parent=4, child=1, left=0, right=10)
        t.sort()
        return t.tree_sequence()

    def test_defaults(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0, 5, 4), tskit.IdentitySegment(5, 10, 4)],
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdPolytomies:
    #
    #          5         |         5
    #         / \        |        / \
    #        4   \       |       4   \
    #       /|\   \      |      /|\   \
    #      / | \   \     |     / | \   \
    #     /  |  \   \    |    /  |  \   \
    #    /   |   \   \   |   /   |   \   \
    #   0    1    2   3  |  0    1    3   2
    #                    |
    #                   0.3

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       1           0
        3       1           0
        4       0           2.5
        5       0           3.5
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0.0     1.0     4       0
        0.0     1.0     4       1
        0.0     0.3     4       2
        0.3     1.0     4       3
        0.3     1.0     5       2
        0.0     0.3     5       3
        0.0     1.0     5       4
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_defaults(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0, 1, 4)],
            (0, 2): [
                tskit.IdentitySegment(0, 0.3, 4),
                tskit.IdentitySegment(0.3, 1, 5),
            ],
            (0, 3): [
                tskit.IdentitySegment(0, 0.3, 5),
                tskit.IdentitySegment(0.3, 1, 4),
            ],
            (1, 2): [
                tskit.IdentitySegment(0, 0.3, 4),
                tskit.IdentitySegment(0.3, 1, 5),
            ],
            (1, 3): [
                tskit.IdentitySegment(0, 0.3, 5),
                tskit.IdentitySegment(0.3, 1, 4),
            ],
            (2, 3): [
                tskit.IdentitySegment(0.3, 1, 5),
                tskit.IdentitySegment(0, 0.3, 5),
            ],
        }
        assert_ibd_equal(ibd_segs, true_segs)

    def test_time(self):
        ibd_segs = ibd_segments(self.ts(), max_time=3)
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0, 1, 4)],
            (0, 2): [tskit.IdentitySegment(0, 0.3, 4)],
            (0, 3): [tskit.IdentitySegment(0.3, 1, 4)],
            (1, 2): [tskit.IdentitySegment(0, 0.3, 4)],
            (1, 3): [tskit.IdentitySegment(0.3, 1, 4)],
        }
        assert_ibd_equal(ibd_segs, true_segs)

    def test_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=0.5)
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0, 1, 4)],
            (0, 2): [tskit.IdentitySegment(0.3, 1, 5)],
            (0, 3): [tskit.IdentitySegment(0.3, 1, 4)],
            (1, 2): [tskit.IdentitySegment(0.3, 1, 5)],
            (1, 3): [tskit.IdentitySegment(0.3, 1, 4)],
            (2, 3): [tskit.IdentitySegment(0.3, 1, 5)],
        }
        (ibd_segs, true_segs)

    def test_input_within(self):
        ibd_segs = ibd_segments(self.ts(), within=[0, 1, 2])
        true_segs = {
            (0, 1): [tskit.IdentitySegment(0, 1, 4)],
            (0, 2): [
                tskit.IdentitySegment(0, 0.3, 4),
                tskit.IdentitySegment(0.3, 1, 5),
            ],
            (1, 2): [
                tskit.IdentitySegment(0, 0.3, 4),
                tskit.IdentitySegment(0.3, 1, 5),
            ],
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdInternalSamples:
    #
    #
    #      3
    #     / \
    #    /   2
    #   /     \
    #  0      (1)

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       0           0
        2       1           1
        3       0           2
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0.0     1.0     2       1
        0.0     1.0     3       0
        0.0     1.0     3       2
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_defaults(self):
        ibd_segs = ibd_segments(self.ts())
        true_segs = {
            (0, 2): [tskit.IdentitySegment(0, 1, 3)],
        }
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdLengthThreshold:
    """
    Tests the behaviour of the min_length argument in niche cases.
    """

    # 2
    #             |     3
    # 1      2    |    / \
    #       / \   |   /   \
    # 0    0   1  |  0     1
    # |------------|----------|
    # 0.0          0.4        1.0

    @tests.cached_example
    def ts(self):
        nodes = io.StringIO(
            """\
        id      is_sample   time
        0       1           0
        1       1           0
        2       0           1
        3       0           1.5
        """
        )
        edges = io.StringIO(
            """\
        left    right   parent  child
        0       0.4     2       0,1
        0.4     1.0     3       0,1
        """
        )
        return tskit.load_text(nodes=nodes, edges=edges, strict=False)

    def test_length_exceeds_segment(self):
        ibd_segs = ibd_segments(self.ts(), min_length=1.1)
        assert_ibd_equal(ibd_segs, {})

    def test_length_is_negative(self):
        with pytest.raises(tskit.LibraryError):
            ibd_segments(self.ts(), min_length=-0.1)

    def test_equal_to_length(self):
        ibd_segs = ibd_segments(self.ts(), min_length=0.4)
        true_segs = {(0, 1): [tskit.IdentitySegment(0.4, 1.0, 3)]}
        assert_ibd_equal(ibd_segs, true_segs)


class TestIbdProperties:
    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_default_within_all_samples(self, ts):
        segs = ts.ibd_segments(store_pairs=True)
        for a, b in segs.keys():
            assert ts.node(a).is_sample()
            assert ts.node(b).is_sample()

    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_within_subset(self, ts):
        samples = ts.samples()
        samples = samples[:3]
        segs = ts.ibd_segments(store_pairs=True, within=samples)
        for a, b in segs.keys():
            assert a in samples
            assert b in samples

    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_between_two_subsets(self, ts):
        samples = ts.samples()
        k = len(samples) // 2
        A = samples[:k]
        B = samples[k:]
        segs = ts.ibd_segments(store_pairs=True, between=[A, B])
        for a, b in segs.keys():
            assert a in A
            assert b in B

    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_between_same_segments_as_filtered_within_pair(self, ts):
        samples = ts.samples()[:10]
        all_segs = ts.ibd_segments(within=samples, store_segments=True)
        A = samples[1::2]
        B = samples[::2]
        between_segs = ts.ibd_segments(store_segments=True, between=[A, B])
        filtered_segs = collections.defaultdict(list)
        for (u, v), seglist in all_segs.items():
            if (u in A and v in B) or (v in A and u in B):
                filtered_segs[(u, v)] = seglist
        assert_ibd_equal(between_segs, filtered_segs)

    @pytest.mark.parametrize("ts", get_example_tree_sequences())
    def test_between_same_segments_as_filtered_within_triple(self, ts):
        samples = ts.samples()[:9]  # Limit the number of samples
        all_segs = ts.ibd_segments(within=samples, store_segments=True)
        A = samples[1::3]
        B = samples[2::3]
        C = samples[0::3]
        all_pairs = set()
        for set_pair in itertools.combinations([A, B, C], 2):
            for pair in itertools.product(*set_pair):
                all_pairs.add(tuple(sorted(pair)))
        between_segs = ts.ibd_segments(store_segments=True, between=[A, B, C])
        filtered_segs = collections.defaultdict(list)
        for pair, seglist in all_segs.items():
            if pair in all_pairs:
                filtered_segs[pair] = seglist
        assert_ibd_equal(between_segs, filtered_segs)


class TestIdentitySegments:
    """
    Test the IdentitySegments class interface.
    """

    def verify_segments(self, ts, ibd_segments):
        samples = set(ts.samples())
        for (a, b), segment_list in ibd_segments.items():
            assert a < b
            assert a in samples
            assert b in samples
            left = segment_list.left
            right = segment_list.right
            node = segment_list.node

            num_segments = 0
            total_span = 0
            for j, seg in enumerate(segment_list):
                assert isinstance(seg, tskit.IdentitySegment)
                total_span += seg.span
                num_segments += 1
                assert seg.span == seg.right - seg.left
                assert seg.left == left[j]
                assert seg.right == right[j]
                assert seg.node == node[j]
                assert 0 <= seg.node < ts.num_nodes
            assert total_span == segment_list.total_span
            assert num_segments == len(segment_list)

        total_span = sum(lst.total_span for lst in ibd_segments.values())
        np.testing.assert_allclose(ibd_segments.total_span, total_span)
        num_segments = sum(len(lst) for lst in ibd_segments.values())
        assert num_segments == ibd_segments.num_segments

    @pytest.mark.parametrize("store_segments", [True, False])
    @pytest.mark.parametrize("store_pairs", [True, False])
    def test_str(self, store_segments, store_pairs):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_segments=store_segments, store_pairs=store_pairs)
        s = str(result)
        assert "IdentitySegments" in s
        assert "max_time" in s
        assert "min_length" in s

    def test_repr_store_segments(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_segments=True)
        s = repr(result)
        assert s.startswith("IdentitySegments({")
        for lst in result.values():
            s = repr(lst)
            assert s.startswith("IdentitySegmentList([")

    def test_repr_without_store_segments(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_pairs=True)
        s = repr(result)
        assert s.startswith("<tskit.tables.IdentitySegments")
        result = ts.ibd_segments()
        s = repr(result)
        assert s.startswith("<tskit.tables.IdentitySegments")

    def test_store_segs_implies_store_pairs(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_pairs=True)
        assert result.num_pairs == 6
        result = ts.ibd_segments(store_segments=True)
        assert result.num_pairs == 6

    def test_operations_available_by_default(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments()
        assert result.num_segments == 6
        assert result.total_span == 6
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = result.num_pairs
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = len(result)
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = result.pairs
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = result[0, 1]
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = list(result)
        with pytest.raises(tskit.IdentityPairsNotStoredError):
            _ = result == result
        # It's OK to when we compare with another type
        assert result != []

    def test_operations_available_store_pairs(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_pairs=True)
        assert result.num_segments == 6
        assert result.total_span == 6
        assert result.num_pairs == 6
        assert len(result) == 6
        assert result.pairs is not None
        seglist = result[0, 1]
        assert seglist.total_span == 1
        assert len(seglist) == 1
        with pytest.raises(tskit.IdentitySegmentsNotStoredError):
            _ = list(seglist)
        with pytest.raises(tskit.IdentitySegmentsNotStoredError):
            _ = seglist.left
        with pytest.raises(tskit.IdentitySegmentsNotStoredError):
            _ = seglist.right
        with pytest.raises(tskit.IdentitySegmentsNotStoredError):
            _ = seglist.node
        with pytest.raises(tskit.IdentitySegmentsNotStoredError):
            _ = seglist == seglist

    @pytest.mark.parametrize("n", [1, 2, 3])
    def test_pairs_all_samples(self, n):
        ts = msprime.sim_ancestry(n, random_seed=2)
        result = ts.ibd_segments(store_segments=True)
        pairs = np.array(list(itertools.combinations(ts.samples(), 2)))
        np.testing.assert_array_equal(pairs, result.pairs)
        self.verify_segments(ts, result)

    @pytest.mark.parametrize("n", [3, 4, 5])
    def test_pairs_subset(self, n):
        ts = msprime.sim_ancestry(n, random_seed=2)
        pairs = np.array([(0, 1), (0, 2), (1, 2)])
        result = ts.ibd_segments(within=[0, 1, 2], store_segments=True)
        np.testing.assert_array_equal(pairs, result.pairs)
        self.verify_segments(ts, result)

    @pytest.mark.parametrize("max_time", [0, 1, 10])
    def test_max_time(self, max_time):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(max_time=max_time, store_segments=True)
        assert result.max_time == max_time
        self.verify_segments(ts, result)

    def test_max_time_default(self):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(store_segments=True)
        assert np.isinf(result.max_time)
        self.verify_segments(ts, result)

    @pytest.mark.parametrize("min_length", [0, 1, 10])
    def test_min_length(self, min_length):
        ts = msprime.sim_ancestry(2, random_seed=2)
        result = ts.ibd_segments(min_length=min_length, store_segments=True)
        assert result.min_length == min_length
        self.verify_segments(ts, result)

    @pytest.mark.parametrize("min_length", [100, 101, 100000])
    def test_min_length_longer_than_seq_length(self, min_length):
        ts = msprime.sim_ancestry(
            100, recombination_rate=0.1, sequence_length=100, random_seed=2
        )
        result = ts.ibd_segments(min_length=min_length, store_segments=True)
        assert result.min_length == min_length
        assert result.num_segments == 0
        self.verify_segments(ts, result)

    def test_recombination_discrete(self):
        ts = msprime.sim_ancestry(
            10, sequence_length=100, recombination_rate=0.1, random_seed=2
        )
        assert ts.num_trees > 2
        result = ts.ibd_segments(store_segments=True)
        self.verify_segments(ts, result)

    def test_recombination_continuous(self):
        ts = msprime.sim_ancestry(
            10,
            recombination_rate=1,
            random_seed=2,
            discrete_genome=False,
            sequence_length=1,
        )
        assert ts.num_trees > 2
        result = ts.ibd_segments(store_segments=True)
        self.verify_segments(ts, result)

    def test_dict_interface(self):
        ts = msprime.sim_ancestry(5, random_seed=2)
        pairs = list(itertools.combinations(ts.samples(), 2))
        result = ts.ibd_segments(store_segments=True)
        assert len(result) == len(pairs)
        for pair in pairs:
            assert pair in result
            assert result[pair] is not None
        for k, v in result.items():
            assert k in pairs
            assert isinstance(v, tskit.IdentitySegmentList)


class TestIdentitySegmentsList:
    """
    Tests for the IdentitySegmentList class.
    """

    example_ts = msprime.sim_ancestry(
        3, sequence_length=100, recombination_rate=0.1, random_seed=2
    )

    def test_list_semantics(self):
        result = self.example_ts.ibd_segments(store_segments=True)
        assert len(result) > 0
        for seglist in result.values():
            lst = list(seglist)
            assert len(lst) == len(seglist)
            assert lst == list(seglist)

    def test_str(self):
        result = self.example_ts.ibd_segments(store_segments=True)
        seglist = list(result.values())[0]
        assert str(seglist).startswith("IdentitySegmentList")

    def test_repr(self):
        result = self.example_ts.ibd_segments(store_segments=True)
        seglist = list(result.values())[0]
        assert repr(seglist).startswith("IdentitySegmentList([IdentitySegment")

    def test_eq_semantics(self):
        result = self.example_ts.ibd_segments(store_segments=True)
        seglists = list(result.values())
        assert len(result) == len(seglists)
        assert len(seglists) > 1
        for seglist1, seglist2 in zip(result.values(), seglists):
            assert seglist1 == seglist2
            assert not (seglist1 != seglist2)
            assert seglist1 != result
            assert seglist1 != []
        # The chance of getting two identical seglists is miniscule
        for seglist in seglists[1:]:
            assert seglist != seglists[0]

    def test_eq_fails_without_store_segments(self):
        result = self.example_ts.ibd_segments(store_pairs=True)
        for seglist in result.values():
            with pytest.raises(tskit.IdentitySegmentsNotStoredError):
                _ = seglist == seglist
            # But it's OK when comparing to another type, since we know
            # it'll be False regardless
            assert seglist != []

    def test_list_contents(self):
        result = self.example_ts.ibd_segments(store_segments=True)
        assert len(result) > 0
        for seglist in result.values():
            for seg in seglist:
                assert isinstance(seg, tskit.IdentitySegment)



# MIT License
#
# Copyright (c) 2020-2021 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Python implementation of the IBD-finding algorithms.
"""
import argparse
import collections

import numpy as np

import tskit


class Segment:
    """
    A class representing a single segment. Each segment has a left and right,
    denoting the loci over which it spans, a node and a next, giving the next
    in the chain.

    The node it records is the *output* node ID.
    """

    def __init__(self, left=None, right=None, node=None, next_seg=None):
        self.left = left
        self.right = right
        self.node = node
        self.next = next_seg

    def __str__(self):
        s = "({}-{}->{}:next={})".format(
            self.left, self.right, self.node, repr(self.next)
        )
        return s

    def __repr__(self):
        return repr((self.left, self.right, self.node))

    def __eq__(self, other):
        # NOTE: to simplify tests, we DON'T check for equality of 'next'.
        return (
            self.left == other.left
            and self.right == other.right
            and self.node == other.node
        )

    def __lt__(self, other):
        return (self.node, self.left, self.right) < (
            other.node,
            other.left,
            other.right,
        )


class SegmentList:
    """
    A class representing a list of segments that are descended from a given ancestral
    node via a particular child of the ancestor.
    Each SegmentList keeps track of the first and last segment in the list, head and
    tail.
    """

    def __init__(self, head=None, tail=None):
        self.head = head
        self.tail = tail

    def __str__(self):
        return repr(self)

    def __repr__(self):
        tuple_segs = []
        seg = self.head
        while seg is not None:
            tuple_segs.append((seg.left, seg.right, seg.node))
            seg = seg.next
        return repr(tuple_segs)

    def extend(self, seglist):
        """
        Extends this segment list with the segments in the specified list.
        """
        assert isinstance(seglist, SegmentList)
        if seglist.head is not None:
            if self.head is None:
                self.head = seglist.head
                self.tail = seglist.tail
            else:
                self.tail.next = seglist.head
                self.tail = seglist.tail

    def append(self, segment):
        """
        Append the specified segment to the end of this list.
        """
        assert isinstance(segment, Segment)
        if self.head is None:
            self.head = segment
            self.tail = segment
        else:
            self.tail.next = segment
            self.tail = segment


class IbdResult:
    """
    Class representing the IBD segments in a tree sequence for a given
    set of sample pairs.
    """

    def __init__(self):
        self.segments = collections.defaultdict(list)

    def __repr__(self):
        return repr(self.segments)

    def __str__(self):
        return repr(self)

    def add_segment(self, a, b, seg):
        key = (a, b) if a < b else (b, a)
        self.segments[key].append(tskit.IdentitySegment(seg.left, seg.right, seg.node))


class IbdFinder:
    """
    Finds all IBD relationships between specified sample pairs in a tree sequence.
    """

    def __init__(self, ts, *, within=None, between=None, min_length=0, max_time=None):
        self.ts = ts
        self.result = IbdResult()
        if within is not None and between is not None:
            raise ValueError("within and between are mutually exclusive")

        self.sample_set_id = np.zeros(ts.num_nodes, dtype=int) - 1
        self.finding_between = False
        if between is not None:
            self.finding_between = True
            for set_id, samples in enumerate(between):
                self.sample_set_id[samples] = set_id
        else:
            if within is None:
                within = ts.samples()
            self.sample_set_id[within] = 0
        self.min_length = min_length
        self.max_time = np.inf if max_time is None else max_time
        self.A = [SegmentList() for _ in range(ts.num_nodes)]  # Descendant segments
        for u in range(ts.num_nodes):
            if self.sample_set_id[u] != -1:
                self.A[u].append(Segment(0, ts.sequence_length, u))
        self.tables = self.ts.tables

    def print_state(self):
        print("IBD Finder")
        print("min_length = ", self.min_length)
        print("max_time   = ", self.max_time)
        print("finding_between = ", self.finding_between)
        print("u\tset_id\tA = ")
        for u, a in enumerate(self.A):
            print(u, self.sample_set_id[u], a, sep="\t")

    def run(self):
        node_times = self.tables.nodes.time
        for e in self.ts.edges():
            time = node_times[e.parent]
            if time > self.max_time:
                # Stop looking for IBD segments once the
                # processed nodes are older than the max time.
                break
            child_segs = SegmentList()
            s = self.A[e.child].head
            while s is not None:
                intvl = (
                    max(e.left, s.left),
                    min(e.right, s.right),
                )
                if intvl[1] - intvl[0] > self.min_length:
                    child_segs.append(Segment(intvl[0], intvl[1], s.node))
                s = s.next
            self.record_ibd(e.parent, child_segs)
            self.A[e.parent].extend(child_segs)
        return self.result.segments

    def record_ibd(self, current_parent, child_segs):
        """
        Given the specified set of child segments for the current parent
        record the IBD segments that will occur as a result of adding these
        new segments into the existing list.
        """
        # Note the implementation here is O(n^2) because we have to compare
        # every segment with every other one. If the segments were stored in
        # left-to-right sorted order, we could avoid and merge them more
        # efficiently. There is some added complexity in doing this, however.
        seg0 = self.A[current_parent].head
        while seg0 is not None:
            seg1 = child_segs.head
            while seg1 is not None:
                left = max(seg0.left, seg1.left)
                right = min(seg0.right, seg1.right)
                # If there are any overlapping segments, record as a new
                # IBD relationship.
                if self.passes_filters(seg0.node, seg1.node, left, right):
                    self.result.add_segment(
                        seg0.node, seg1.node, Segment(left, right, current_parent)
                    )
                seg1 = seg1.next
            seg0 = seg0.next

    def passes_filters(self, a, b, left, right):
        if a == b:
            return False
        if right - left <= self.min_length:
            return False
        if self.finding_between:
            return self.sample_set_id[a] != self.sample_set_id[b]
        else:
            return True


if __name__ == "__main__":
    """
    A simple CLI for running IBDFinder on a command line from the `python`
    subdirectory. Basic usage:
    > python3 ./tests/ibd.py --infile test.trees
    """

    parser = argparse.ArgumentParser(
        description="Command line interface for the IBDFinder."
    )

    parser.add_argument(
        "--infile",
        type=str,
        dest="infile",
        nargs=1,
        metavar="IN_FILE",
        help="The tree sequence to be analysed.",
    )

    parser.add_argument(
        "--min-length",
        type=float,
        dest="min_length",
        nargs=1,
        metavar="MIN_LENGTH",
        help="Only segments longer than this cutoff will be returned.",
    )

    parser.add_argument(
        "--max-time",
        type=float,
        dest="max_time",
        nargs=1,
        metavar="MAX_TIME",
        help="Only segments younger this time will be returned.",
    )

    parser.add_argument(
        "--samples",
        type=int,
        dest="samples",
        nargs=2,
        metavar="SAMPLES",
        help="If provided, only this pair's IBD info is returned.",
    )

    args = parser.parse_args()
    ts = tskit.load(args.infile[0])
    if args.min_length is None:
        min_length = 0
    else:
        min_length = args.min_length[0]
    if args.max_time is None:
        max_time = None
    else:
        max_time = args.max_time[0]

    s = IbdFinder(ts, samples=ts.samples(), min_length=min_length, max_time=max_time)
    all_segs = s.find_ibd_segments()

    if args.samples is None:
        print(all_segs)
    else:
        samples = args.samples
        print(all_segs[(samples[0], samples[1])])



# MIT License
#
# Copyright (c) 2018-2020 Tskit Developers
# Copyright (c) 2015-2018 University of Oxford
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Python implementation of the simplify algorithm.
"""
import sys

import numpy as np
import portion

import tskit


def overlapping_segments(segments):
    """
    Returns an iterator over the (left, right, X) tuples describing the
    distinct overlapping segments in the specified set.
    """
    S = sorted(segments, key=lambda x: x.left)
    n = len(S)
    # Insert a sentinel at the end for convenience.
    S.append(Segment(sys.float_info.max, 0))
    right = S[0].left
    X = []
    j = 0
    while j < n:
        # Remove any elements of X with right <= left
        left = right
        X = [x for x in X if x.right > left]
        if len(X) == 0:
            left = S[j].left
        while j < n and S[j].left == left:
            X.append(S[j])
            j += 1
        j -= 1
        right = min(x.right for x in X)
        right = min(right, S[j + 1].left)
        yield left, right, X
        j += 1

    while len(X) > 0:
        left = right
        X = [x for x in X if x.right > left]
        if len(X) > 0:
            right = min(x.right for x in X)
            yield left, right, X


class Segment:
    """
    A class representing a single segment. Each segment has a left and right,
    denoting the loci over which it spans, a node and a next, giving the next
    in the chain.

    The node it records is the *output* node ID.
    """

    def __init__(self, left=None, right=None, node=None, next_segment=None):
        self.left = left
        self.right = right
        self.node = node
        self.next = next_segment

    def __str__(self):
        s = "({}-{}->{}:next={})".format(
            self.left, self.right, self.node, repr(self.next)
        )
        return s

    def __repr__(self):
        return repr((self.left, self.right, self.node))

    def __lt__(self, other):
        return (self.left, self.right, self.node) < (other.left, other.right, self.node)


class Simplifier:
    """
    Simplifies a tree sequence to its minimal representation given a subset
    of the leaves.
    """

    def __init__(
        self,
        ts,
        sample,
        reduce_to_site_topology=False,
        filter_sites=True,
        filter_populations=True,
        filter_individuals=True,
        keep_unary=False,
        keep_unary_in_individuals=False,
        keep_input_roots=False,
    ):
        self.ts = ts
        self.n = len(sample)
        self.reduce_to_site_topology = reduce_to_site_topology
        self.sequence_length = ts.sequence_length
        self.filter_sites = filter_sites
        self.filter_populations = filter_populations
        self.filter_individuals = filter_individuals
        self.keep_unary = keep_unary
        self.keep_unary_in_individuals = keep_unary_in_individuals
        self.keep_input_roots = keep_input_roots
        self.num_mutations = ts.num_mutations
        self.input_sites = list(ts.sites())
        self.A_head = [None for _ in range(ts.num_nodes)]
        self.A_tail = [None for _ in range(ts.num_nodes)]
        self.tables = self.ts.tables.copy()
        self.tables.clear()
        self.edge_buffer = {}
        self.node_id_map = np.zeros(ts.num_nodes, dtype=np.int32) - 1
        self.mutation_node_map = [-1 for _ in range(self.num_mutations)]
        self.samples = set(sample)
        self.sort_offset = -1
        # We keep a map of input nodes to mutations.
        self.mutation_map = [[] for _ in range(ts.num_nodes)]
        position = ts.tables.sites.position
        site = ts.tables.mutations.site
        node = ts.tables.mutations.node
        for mutation_id in range(ts.num_mutations):
            site_position = position[site[mutation_id]]
            self.mutation_map[node[mutation_id]].append((site_position, mutation_id))
        for sample_id in sample:
            output_id = self.record_node(sample_id, is_sample=True)
            self.add_ancestry(sample_id, 0, self.sequence_length, output_id)
        self.position_lookup = None
        if self.reduce_to_site_topology:
            self.position_lookup = np.hstack([[0], position, [self.sequence_length]])

    def record_node(self, input_id, is_sample=False):
        """
        Adds a new node to the output table corresponding to the specified input
        node ID.
        """
        node = self.ts.node(input_id)
        flags = node.flags
        # Need to zero out the sample flag
        flags &= ~tskit.NODE_IS_SAMPLE
        if is_sample:
            flags |= tskit.NODE_IS_SAMPLE
        output_id = self.tables.nodes.append(node.replace(flags=flags))
        self.node_id_map[input_id] = output_id
        return output_id

    def rewind_node(self, input_id, output_id):
        """
        Remove the mapping for the specified input and output node pair. This is
        done because there are no edges referring to the node.
        """
        assert output_id == len(self.tables.nodes) - 1
        assert output_id == self.node_id_map[input_id]
        self.tables.nodes.truncate(output_id)
        self.node_id_map[input_id] = -1

    def flush_edges(self):
        """
        Flush the edges to the output table after sorting and squashing
        any redundant records.
        """
        num_edges = 0
        for child in sorted(self.edge_buffer.keys()):
            for edge in self.edge_buffer[child]:
                self.tables.edges.append(edge)
                num_edges += 1
        self.edge_buffer.clear()
        return num_edges

    def record_edge(self, left, right, parent, child):
        """
        Adds an edge to the output list.
        """
        if self.reduce_to_site_topology:
            X = self.position_lookup
            left_index = np.searchsorted(X, left)
            right_index = np.searchsorted(X, right)
            # Find the smallest site position index greater than or equal to left
            # and right, i.e., slide each endpoint of an interval to the right
            # until they hit a site position. If both left and right map to the
            # the same position then we discard this edge. We also discard an
            # edge if left = 0 and right is less than the first site position.
            if left_index == right_index or (left_index == 0 and right_index == 1):
                return
            # Remap back to zero if the left end maps to the first site.
            if left_index == 1:
                left_index = 0
            left = X[left_index]
            right = X[right_index]
        if child not in self.edge_buffer:
            self.edge_buffer[child] = [tskit.Edge(left, right, parent, child)]
        else:
            last = self.edge_buffer[child][-1]
            if last.right == left:
                last.right = right
            else:
                self.edge_buffer[child].append(tskit.Edge(left, right, parent, child))

    def print_state(self):
        print(".................")
        print("Ancestors: ")
        num_nodes = len(self.A_tail)
        for j in range(num_nodes):
            print("\t", j, "->", end="")
            x = self.A_head[j]
            while x is not None:
                print(f"({x.left}-{x.right}->{x.node})", end="")
                x = x.next
            print()
        print("Mutation map:")
        for u in range(len(self.mutation_map)):
            v = self.mutation_map[u]
            if len(v) > 0:
                print("\t", u, "->", v)
        print("Node ID map: (input->output)")
        for input_id, output_id in enumerate(self.node_id_map):
            print("\t", input_id, "->", output_id)
        print("Mutation node map")
        for j in range(self.num_mutations):
            print("\t", j, "->", self.mutation_node_map[j])
        print("Output:")
        print(self.tables)
        self.check_state()

    def map_mutations(self, left, right, input_id, output_id):
        """
        Map any mutations for the input node ID on the
        interval to its output ID.
        """
        assert output_id != -1
        # TODO we should probably remove these as they are used.
        # Or else, binary search the list so it's quick.
        for x, mutation_id in self.mutation_map[input_id]:
            if left <= x < right:
                self.mutation_node_map[mutation_id] = output_id

    def add_ancestry(self, input_id, left, right, node):
        tail = self.A_tail[input_id]
        if tail is None:
            x = Segment(left, right, node)
            self.A_head[input_id] = x
            self.A_tail[input_id] = x
        else:
            if tail.right == left and tail.node == node:
                tail.right = right
            else:
                x = Segment(left, right, node)
                tail.next = x
                self.A_tail[input_id] = x

        self.map_mutations(left, right, input_id, node)

    def merge_labeled_ancestors(self, S, input_id):
        """
        All ancestry segments in S come together into a new parent.
        The new parent must be assigned and any overlapping segments coalesced.
        """
        output_id = self.node_id_map[input_id]
        is_sample = output_id != -1
        if is_sample:
            # Free up the existing ancestry mapping.
            x = self.A_tail[input_id]
            assert x.left == 0 and x.right == self.sequence_length
            self.A_tail[input_id] = None
            self.A_head[input_id] = None

        prev_right = 0
        for left, right, X in overlapping_segments(S):
            if len(X) == 1:
                ancestry_node = X[0].node
                if is_sample:
                    self.record_edge(left, right, output_id, ancestry_node)
                    ancestry_node = output_id
                elif self.keep_unary or (
                    self.keep_unary_in_individuals
                    and self.ts.node(input_id).individual >= 0
                ):
                    if output_id == -1:
                        output_id = self.record_node(input_id)
                    self.record_edge(left, right, output_id, ancestry_node)
            else:
                if output_id == -1:
                    output_id = self.record_node(input_id)
                ancestry_node = output_id
                for x in X:
                    self.record_edge(left, right, output_id, x.node)
            if is_sample and left != prev_right:
                # Fill in any gaps in the ancestry for the sample
                self.add_ancestry(input_id, prev_right, left, output_id)
            if self.keep_unary or (
                self.keep_unary_in_individuals
                and self.ts.node(input_id).individual >= 0
            ):
                ancestry_node = output_id
            self.add_ancestry(input_id, left, right, ancestry_node)
            prev_right = right

        if is_sample and prev_right != self.sequence_length:
            # If a trailing gap exists in the sample ancestry, fill it in.
            self.add_ancestry(input_id, prev_right, self.sequence_length, output_id)
        if output_id != -1:
            num_edges = self.flush_edges()
            if num_edges == 0 and not is_sample:
                self.rewind_node(input_id, output_id)

    def extract_ancestry(self, edge):
        S = []
        x = self.A_head[edge.child]

        x_head = None
        x_prev = None
        while x is not None:
            if x.right > edge.left and edge.right > x.left:
                y = Segment(max(x.left, edge.left), min(x.right, edge.right), x.node)
                # print("snip", y)
                S.append(y)
                assert x.left <= y.left
                assert x.right >= y.right
                seg_left = None
                seg_right = None
                if x.left != y.left:
                    seg_left = Segment(x.left, y.left, x.node)
                    if x_prev is None:
                        x_head = seg_left
                    else:
                        x_prev.next = seg_left
                    x_prev = seg_left
                if x.right != y.right:
                    x.left = y.right
                    seg_right = x
                else:
                    # Free x
                    seg_right = x.next
                if x_prev is None:
                    x_head = seg_right
                else:
                    x_prev.next = seg_right
                x = seg_right
            else:
                if x_prev is None:
                    x_head = x
                x_prev = x
                x = x.next
        # Note - we had some code to defragment segments in the output
        # chain here, but couldn't find an example where it needed to
        # be called. So, looks like squashing isn't necessary here.
        self.A_head[edge.child] = x_head
        self.A_tail[edge.child] = x_prev
        return S

    def process_parent_edges(self, edges):
        """
        Process all of the edges for a given parent.
        """
        assert len({e.parent for e in edges}) == 1
        parent = edges[0].parent
        S = []
        for edge in edges:
            S.extend(self.extract_ancestry(edge))
        self.merge_labeled_ancestors(S, parent)
        self.check_state()

    def finalise_sites(self):
        # Build a map from the old mutation IDs to new IDs. Any mutation that
        # has not been mapped to a node in the new tree sequence will be removed.
        mutation_id_map = [-1 for _ in range(self.num_mutations)]
        num_output_mutations = 0

        for site in self.ts.sites():
            num_output_site_mutations = 0
            for mut in site.mutations:
                mapped_node = self.mutation_node_map[mut.id]
                mapped_parent = -1
                if mut.parent != -1:
                    mapped_parent = mutation_id_map[mut.parent]
                if mapped_node != -1:
                    mutation_id_map[mut.id] = num_output_mutations
                    num_output_mutations += 1
                    num_output_site_mutations += 1
            output_site = True
            if self.filter_sites and num_output_site_mutations == 0:
                output_site = False

            if output_site:
                for mut in site.mutations:
                    if mutation_id_map[mut.id] != -1:
                        mapped_parent = -1
                        if mut.parent != -1:
                            mapped_parent = mutation_id_map[mut.parent]
                        self.tables.mutations.append(
                            mut.replace(
                                site=len(self.tables.sites),
                                node=self.mutation_node_map[mut.id],
                                parent=mapped_parent,
                            )
                        )
                self.tables.sites.append(site)

    def finalise_references(self):
        input_populations = self.ts.tables.populations
        population_id_map = np.arange(len(input_populations) + 1, dtype=np.int32)
        # Trick to ensure the null population gets mapped to itself.
        population_id_map[-1] = -1
        input_individuals = self.ts.tables.individuals
        individual_id_map = np.arange(len(input_individuals) + 1, dtype=np.int32)
        # Trick to ensure the null individual gets mapped to itself.
        individual_id_map[-1] = -1

        population_ref_count = np.ones(len(input_populations), dtype=int)
        if self.filter_populations:
            population_ref_count[:] = 0
            population_id_map[:] = -1
        individual_ref_count = np.ones(len(input_individuals), dtype=int)
        if self.filter_individuals:
            individual_ref_count[:] = 0
            individual_id_map[:] = -1

        for node in self.tables.nodes:
            if self.filter_populations and node.population != tskit.NULL:
                population_ref_count[node.population] += 1
            if self.filter_individuals and node.individual != tskit.NULL:
                individual_ref_count[node.individual] += 1

        for input_id, count in enumerate(population_ref_count):
            if count > 0:
                row = input_populations[input_id]
                output_id = self.tables.populations.append(row)
                population_id_map[input_id] = output_id
        for input_id, count in enumerate(individual_ref_count):
            if count > 0:
                row = input_individuals[input_id]
                output_id = self.tables.individuals.append(row)
                individual_id_map[input_id] = output_id

        # Remap the population ID references for nodes.
        nodes = self.tables.nodes
        nodes.set_columns(
            flags=nodes.flags,
            time=nodes.time,
            metadata=nodes.metadata,
            metadata_offset=nodes.metadata_offset,
            individual=individual_id_map[nodes.individual],
            population=population_id_map[nodes.population],
        )

        # Remap the parent ids of individuals
        individuals_copy = self.tables.individuals.copy()
        self.tables.individuals.clear()
        for row in individuals_copy:
            mapped_parents = []
            for p in row.parents:
                if p == -1:
                    mapped_parents.append(-1)
                else:
                    mapped_parents.append(individual_id_map[p])
            self.tables.individuals.append(row.replace(parents=mapped_parents))

        # We don't support migrations for now. We'll need to remap these as well.
        assert self.ts.num_migrations == 0

    def insert_input_roots(self):
        youngest_root_time = np.inf
        for input_id in range(len(self.node_id_map)):
            x = self.A_head[input_id]
            if x is not None:
                output_id = self.node_id_map[input_id]
                if output_id == -1:
                    output_id = self.record_node(input_id)
                while x is not None:
                    if x.node != output_id:
                        self.record_edge(x.left, x.right, output_id, x.node)
                        self.map_mutations(x.left, x.right, input_id, output_id)
                    x = x.next
                self.flush_edges()
                root_time = self.tables.nodes.time[output_id]
                if root_time < youngest_root_time:
                    youngest_root_time = root_time
        # We have to sort the edge table from the point where the edges
        # for the youngest root would be inserted.
        # Note: it would be nicer to do the sort here, but we have to
        # wait until the finalise_references method has been called to
        # make sure all the populations etc have been setup.
        node_time = self.tables.nodes.time
        edge_parent = self.tables.edges.parent
        offset = 0
        while (
            offset < len(self.tables.edges)
            and node_time[edge_parent[offset]] < youngest_root_time
        ):
            offset += 1
        self.sort_offset = offset

    def simplify(self):
        if self.ts.num_edges > 0:
            all_edges = list(self.ts.edges())
            edges = all_edges[:1]
            for e in all_edges[1:]:
                if e.parent != edges[0].parent:
                    self.process_parent_edges(edges)
                    edges = []
                edges.append(e)
            self.process_parent_edges(edges)
        if self.keep_input_roots:
            self.insert_input_roots()
        self.finalise_sites()
        self.finalise_references()
        if self.sort_offset != -1:
            self.tables.sort(edge_start=self.sort_offset)
        ts = self.tables.tree_sequence()
        return ts, self.node_id_map

    def check_state(self):
        # print("CHECK_STATE")
        all_ancestry = []
        num_nodes = len(self.A_head)
        for j in range(num_nodes):
            head = self.A_head[j]
            tail = self.A_tail[j]
            if head is None:
                assert tail is None
            else:
                x = head
                while x.next is not None:
                    assert x.right <= x.next.left
                    x = x.next
                assert x == tail
                x = head
                while x is not None:
                    assert x.left < x.right
                    all_ancestry.append(portion.openclosed(x.left, x.right))
                    if x.next is not None:
                        assert x.right <= x.next.left
                        # We should also not have any squashable segments.
                        if x.right == x.next.left:
                            assert x.node != x.next.node
                    x = x.next
        # Make sure we haven't lost ancestry.
        if len(all_ancestry) > 0:
            union = all_ancestry[0]
            for interval in all_ancestry[1:]:
                union = union.union(interval)
            assert union.atomic
            assert union.lower == 0
            assert union.upper == self.sequence_length


class AncestorMap:
    """
    Simplifies a tree sequence to show relationships between
    samples and a designated set of ancestors.
    """

    def __init__(self, ts, sample, ancestors):
        self.ts = ts
        self.samples = set(sample)
        assert (self.samples).issubset(set(range(0, ts.num_nodes)))
        self.ancestors = set(ancestors)
        assert (self.ancestors).issubset(set(range(0, ts.num_nodes)))
        self.table = tskit.EdgeTable()
        self.sequence_length = ts.sequence_length
        self.A_head = [None for _ in range(ts.num_nodes)]
        self.A_tail = [None for _ in range(ts.num_nodes)]
        for sample_id in sample:
            self.add_ancestry(0, self.sequence_length, sample_id, sample_id)
        self.edge_buffer = {}

    def link_ancestors(self):
        if self.ts.num_edges > 0:
            all_edges = list(self.ts.edges())
            edges = all_edges[:1]
            for e in all_edges[1:]:
                if e.parent != edges[0].parent:
                    self.process_parent_edges(edges)
                    edges = []
                edges.append(e)
            self.process_parent_edges(edges)
        return self.table

    def process_parent_edges(self, edges):
        """
        Process all of the edges for a given parent.
        """
        assert len({e.parent for e in edges}) == 1
        parent = edges[0].parent
        S = []
        for edge in edges:
            x = self.A_head[edge.child]
            while x is not None:
                if x.right > edge.left and edge.right > x.left:
                    y = Segment(
                        max(x.left, edge.left), min(x.right, edge.right), x.node
                    )
                    S.append(y)
                x = x.next
        self.merge_labeled_ancestors(S, parent)
        self.check_state()

    def merge_labeled_ancestors(self, S, input_id):
        """
        All ancestry segments in S come together into a new parent.
        The new parent must be assigned and any overlapping segments coalesced.
        """
        is_sample = input_id in self.samples
        if is_sample:
            # Free up the existing ancestry mapping.
            x = self.A_tail[input_id]
            assert x.left == 0 and x.right == self.sequence_length
            self.A_tail[input_id] = None
            self.A_head[input_id] = None

        is_ancestor = input_id in self.ancestors
        prev_right = 0
        for left, right, X in overlapping_segments(S):
            if is_ancestor or is_sample:
                for x in X:
                    ancestry_node = x.node
                    self.record_edge(left, right, input_id, ancestry_node)
                self.add_ancestry(left, right, input_id, input_id)

                if is_sample and left != prev_right:
                    # Fill in any gaps in the ancestry for the sample.
                    self.add_ancestry(prev_right, left, input_id, input_id)

            else:
                for x in X:
                    ancestry_node = x.node
                    # Add sample ancestry for the currently-processed segment set.
                    self.add_ancestry(left, right, ancestry_node, input_id)
            prev_right = right

        if is_sample and prev_right != self.sequence_length:
            # If a trailing gap exists in the sample ancestry, fill it in.
            self.add_ancestry(prev_right, self.sequence_length, input_id, input_id)
        if input_id != -1:
            self.flush_edges()

    def record_edge(self, left, right, parent, child):
        """
        Adds an edge to the output list.
        """
        if child not in self.edge_buffer:
            self.edge_buffer[child] = [tskit.Edge(left, right, parent, child)]
        else:
            last = self.edge_buffer[child][-1]
            if last.right == left:
                last.right = right
            else:
                self.edge_buffer[child].append(tskit.Edge(left, right, parent, child))

    def add_ancestry(self, left, right, node, current_node):
        tail = self.A_tail[current_node]
        if tail is None:
            x = Segment(left, right, node)
            self.A_head[current_node] = x
            self.A_tail[current_node] = x
        else:
            if tail.right == left and tail.node == node:
                tail.right = right
            else:
                x = Segment(left, right, node)
                tail.next = x
                self.A_tail[current_node] = x

    def flush_edges(self):
        """
        Flush the edges to the output table after sorting and squashing
        any redundant records.
        """
        num_edges = 0
        for child in sorted(self.edge_buffer.keys()):
            for edge in self.edge_buffer[child]:
                self.table.append(edge)
                num_edges += 1
        self.edge_buffer.clear()
        return num_edges

    def check_state(self):

        num_nodes = len(self.A_head)
        for j in range(num_nodes):
            head = self.A_head[j]
            tail = self.A_tail[j]
            if head is None:
                assert tail is None
            else:
                x = head
                while x.next is not None:
                    x = x.next
                assert x == tail
                x = head.next
                while x is not None:
                    assert x.left < x.right
                    if x.next is not None:
                        if self.ancestors is None:
                            assert x.right <= x.next.left
                        # We should also not have any squashable segments.
                        if x.right == x.next.left:
                            assert x.node != x.next.node
                    x = x.next

    def print_state(self):
        print(".................")
        print("Ancestors: ")
        num_nodes = len(self.A_tail)
        for j in range(num_nodes):
            print("\t", j, "->", end="")
            x = self.A_head[j]
            while x is not None:
                print(f"({x.left}-{x.right}->{x.node})", end="")
                x = x.next
            print()
        print("Output:")
        print(self.table)
        self.check_state()


if __name__ == "__main__":
    # Simple CLI for running simplifier/ancestor mapping above.
    class_to_implement = sys.argv[1]
    assert class_to_implement == "Simplifier" or class_to_implement == "AncestorMap"
    ts = tskit.load(sys.argv[2])

    if class_to_implement == "Simplifier":

        samples = list(map(int, sys.argv[3:]))

        print("When keep_unary = True:")
        s = Simplifier(ts, samples, keep_unary=True)
        # s.print_state()
        tss, _ = s.simplify()
        tables = tss.dump_tables()
        print(tables.nodes)
        print(tables.edges)
        print(tables.sites)
        print(tables.mutations)

        print("\nWhen keep_unary = False")
        s = Simplifier(ts, samples, keep_unary=False)
        # s.print_state()
        tss, _ = s.simplify()
        tables = tss.dump_tables()
        print(tables.nodes)
        print(tables.edges)
        print(tables.sites)
        print(tables.mutations)

        print("\nWhen keep_unary_in_individuals = True")
        s = Simplifier(ts, samples, keep_unary_in_individuals=True)
        # s.print_state()
        tss, _ = s.simplify()
        tables = tss.dump_tables()
        print(tables.nodes)
        print(tables.edges)
        print(tables.sites)
        print(tables.mutations)

    elif class_to_implement == "AncestorMap":

        samples = sys.argv[3]
        samples = samples.split(",")
        samples = list(map(int, samples))

        ancestors = sys.argv[4]
        ancestors = ancestors.split(",")
        ancestors = list(map(int, ancestors))

        s = AncestorMap(ts, samples, ancestors)
        tss = s.link_ancestors()
        # tables = tss.dump_tables()
        # print(tables.nodes)
        print(tss)



# MIT License
#
# Copyright (c) 2018-2021 Tskit Developers
# Copyright (C) 2016 University of Oxford
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Test cases for stats calculations in tskit.
"""
import contextlib
import io

import msprime
import numpy as np
import pytest

import _tskit
import tests
import tests.test_wright_fisher as wf
import tests.tsutil as tsutil
import tskit


@contextlib.contextmanager
def suppress_division_by_zero_warning():
    with np.errstate(invalid="ignore", divide="ignore"):
        yield


def get_r2_matrix(ts):
    """
    Simple site-based version assuming biallic sites.
    """
    A = np.zeros((ts.num_sites, ts.num_sites))
    G = ts.genotype_matrix()
    n = ts.num_samples
    for a in range(ts.num_sites):
        A[a, a] = 1
        fA = np.sum(G[a] != 0) / n
        for b in range(a + 1, ts.num_sites):
            fB = np.sum(G[b] != 0) / n
            nAB = np.sum(np.logical_and(G[a] != 0, G[b] != 0))
            fAB = nAB / n
            D = fAB - fA * fB
            denom = fA * fB * (1 - fA) * (1 - fB)
            A[a, b] = D * D
            with suppress_division_by_zero_warning():
                A[a, b] /= denom
            A[b, a] = A[a, b]
    return A


def _compute_r2(tree, n, f_a, site_b):
    assert len(site_b.mutations) == 1
    assert site_b.ancestral_state != site_b.mutations[0].derived_state
    f_b = tree.num_samples(site_b.mutations[0].node) / n
    f_ab = tree.num_tracked_samples(site_b.mutations[0].node) / n
    D2 = (f_ab - f_a * f_b) ** 2
    denom = f_a * f_b * (1 - f_a) * (1 - f_b)
    if denom == 0:
        return np.nan
    return D2 / denom


def ts_r2(ts, a, b):
    """
    Returns the r2 value between sites a and b in the specified tree sequence.
    """
    a, b = (a, b) if a < b else (b, a)
    site_a = ts.site(a)
    site_b = ts.site(b)
    assert len(site_a.mutations) == 1
    assert len(site_b.mutations) == 1
    n = ts.num_samples
    tree = ts.at(site_a.position)
    a_samples = list(tree.samples(site_a.mutations[0].node))
    f_a = len(a_samples) / n
    tree = ts.at(site_b.position, tracked_samples=a_samples)
    return _compute_r2(tree, n, f_a, site_b)


class LdArrayCalculator:
    """
    Utility class to help organise the state required when tracking all
    the different termination conditions.
    """

    def __init__(self, ts, focal_site_id, direction, max_sites, max_distance):
        self.ts = ts
        self.focal_site = ts.site(focal_site_id)
        self.direction = direction
        self.max_sites = max_sites
        self.max_distance = max_distance
        self.result = []
        self.tree = None

    def _check_site(self, site):
        assert len(site.mutations) == 1
        assert site.ancestral_state != site.mutations[0].derived_state

    def _compute_and_append(self, target_site):
        self._check_site(target_site)

        distance = abs(target_site.position - self.focal_site.position)
        if distance > self.max_distance or len(self.result) >= self.max_sites:
            return True
        r2 = _compute_r2(
            self.tree, self.ts.num_samples, self.focal_frequency, target_site
        )
        self.result.append(r2)
        return False

    def _compute_forward(self):
        done = False
        for site in self.tree.sites():
            if site.id > self.focal_site.id:
                done = self._compute_and_append(site)
                if done:
                    break
        while self.tree.next() and not done:
            for site in self.tree.sites():
                done = self._compute_and_append(site)
                if done:
                    break

    def _compute_backward(self):
        done = False
        for site in reversed(list(self.tree.sites())):
            if site.id < self.focal_site.id:
                done = self._compute_and_append(site)
                if done:
                    break
        while self.tree.prev() and not done:
            for site in reversed(list(self.tree.sites())):
                done = self._compute_and_append(site)
                if done:
                    break

    def run(self):
        self._check_site(self.focal_site)
        self.tree = self.ts.at(self.focal_site.position)
        a_samples = list(self.tree.samples(self.focal_site.mutations[0].node))
        self.focal_frequency = len(a_samples) / self.ts.num_samples

        # Now set the tracked samples on the tree. We don't have a python
        # API for doing this, so we just create a new tree.
        self.tree = self.ts.at(self.focal_site.position, tracked_samples=a_samples)
        if self.direction == 1:
            self._compute_forward()
        else:
            self._compute_backward()
        return np.array(self.result)


def ts_r2_array(ts, a, *, direction=1, max_sites=None, max_distance=None):
    max_sites = ts.num_sites if max_sites is None else max_sites
    max_distance = np.inf if max_distance is None else max_distance
    calc = LdArrayCalculator(ts, a, direction, max_sites, max_distance)
    return calc.run()


class TestLdSingleTree:
    # 2.00┊   4   ┊
    #     ┊ ┏━┻┓  ┊
    # 1.00┊ ┃  3  ┊
    #     ┊ ┃ ┏┻┓ ┊
    # 0.00┊ 0 1 2 ┊
    #     0      10
    #      | |  |
    #  pos 2 4  9
    # node 0 1  0
    @tests.cached_example
    def ts(self):
        ts = tskit.Tree.generate_balanced(3, span=10).tree_sequence
        tables = ts.dump_tables()
        tables.sites.add_row(2, ancestral_state="A")
        tables.sites.add_row(4, ancestral_state="A")
        tables.sites.add_row(9, ancestral_state="T")
        tables.mutations.add_row(site=0, node=0, derived_state="G")
        tables.mutations.add_row(site=1, node=3, derived_state="C")
        tables.mutations.add_row(site=2, node=0, derived_state="G")
        return tables.tree_sequence()

    @pytest.mark.parametrize(["a", "b", "expected"], [(0, 0, 1), (0, 1, 1), (0, 2, 1)])
    def test_r2(self, a, b, expected):
        ts = self.ts()
        A = get_r2_matrix(ts)
        ldc = tskit.LdCalculator(ts)
        assert ldc.r2(a, b) == pytest.approx(expected)
        assert ts_r2(ts, a, b) == pytest.approx(expected)
        assert A[a, b] == pytest.approx(expected)
        assert ldc.r2(b, a) == pytest.approx(expected)
        assert ts_r2(ts, b, a) == pytest.approx(expected)
        assert A[b, a] == pytest.approx(expected)

    @pytest.mark.parametrize("a", [0, 1, 2])
    @pytest.mark.parametrize("direction", [1, -1])
    def test_r2_array(self, a, direction):
        ts = self.ts()
        ldc = tskit.LdCalculator(ts)
        lib_a = ldc.r2_array(a, direction=direction)
        py_a = ts_r2_array(ts, a, direction=direction)
        np.testing.assert_array_almost_equal(lib_a, py_a)


class TestLdFixedSites:
    # 2.00┊   4   ┊
    #     ┊ ┏━┻┓  ┊
    # 1.00┊ ┃  3  ┊
    #     ┊ ┃ ┏┻┓ ┊
    # 0.00┊ 0 1 2 ┊
    #     0      10
    #      | |  |
    #  pos 2 4  9
    # node 0 1  0
    @tests.cached_example
    def ts(self):
        ts = tskit.Tree.generate_balanced(3, span=10).tree_sequence
        tables = ts.dump_tables()
        # First and last mutations are over the root
        tables.sites.add_row(2, ancestral_state="A")
        tables.sites.add_row(4, ancestral_state="A")
        tables.sites.add_row(9, ancestral_state="T")
        tables.mutations.add_row(site=0, node=4, derived_state="G")
        tables.mutations.add_row(site=1, node=3, derived_state="C")
        tables.mutations.add_row(site=2, node=4, derived_state="G")
        return tables.tree_sequence()

    def test_r2_fixed_fixed(self):
        ts = self.ts()
        A = get_r2_matrix(ts)
        ldc = tskit.LdCalculator(ts)
        assert np.isnan(ldc.r2(0, 2))
        assert np.isnan(ts_r2(ts, 0, 2))
        assert np.isnan(A[0, 2])

    def test_r2_fixed_non_fixed(self):
        ts = self.ts()
        A = get_r2_matrix(ts)
        ldc = tskit.LdCalculator(ts)
        assert np.isnan(ldc.r2(0, 1))
        assert np.isnan(ts_r2(ts, 0, 1))
        assert np.isnan(A[0, 1])

    def test_r2_non_fixed_fixed(self):
        ts = self.ts()
        A = get_r2_matrix(ts)
        ldc = tskit.LdCalculator(ts)
        assert np.isnan(ldc.r2(1, 0))
        assert np.isnan(ts_r2(ts, 1, 0))
        assert np.isnan(A[1, 0])


class BaseTestLd:
    """
    Define a set of tests for LD calculations. Subclasses should be
    concrete examples with at least two sites which implement a
    method ts() which returns the tree sequence and the full LD
    matrix.
    """

    def test_r2_all_pairs(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        for j in range(ts.num_sites):
            for k in range(ts.num_sites):
                r2 = A[j, k]
                assert ldc.r2(j, k) == pytest.approx(r2)
                assert ts_r2(ts, j, k) == pytest.approx(r2)

    def test_r2_array_first_site_forward(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        A1 = ldc.r2_array(0, direction=1)
        A2 = ts_r2_array(ts, 0, direction=1)
        np.testing.assert_array_almost_equal(A2, A[0, 1:])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_mid_forward(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        site = ts.num_sites // 2
        A1 = ldc.r2_array(site, direction=1)
        A2 = ts_r2_array(ts, site, direction=1)
        np.testing.assert_array_almost_equal(A2, A[site, site + 1 :])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_first_site_forward_max_sites(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        A1 = ldc.r2_array(0, direction=1, max_sites=2)
        A2 = ts_r2_array(ts, 0, direction=1, max_sites=2)
        np.testing.assert_array_almost_equal(A2, A[0, 1:3])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_first_site_forward_max_distance(self):
        ts, _ = self.ts()
        ldc = tskit.LdCalculator(ts)
        A1 = ldc.r2_array(0, direction=1, max_distance=3)
        A2 = ts_r2_array(ts, 0, direction=1, max_distance=3)
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_last_site_backward(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        a = ts.num_sites - 1
        A1 = ldc.r2_array(a, direction=-1)
        A2 = ts_r2_array(ts, a, direction=-1)
        np.testing.assert_array_almost_equal(A2, A[-1, :-1][::-1])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_mid_backward(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        site = ts.num_sites // 2
        A1 = ldc.r2_array(site, direction=-1)
        A2 = ts_r2_array(ts, site, direction=-1)
        np.testing.assert_array_almost_equal(A2, A[site, :site][::-1])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_last_site_backward_max_sites(self):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        a = ts.num_sites - 1
        A1 = ldc.r2_array(a, direction=-1, max_sites=2)
        A2 = ts_r2_array(ts, a, direction=-1, max_sites=2)
        np.testing.assert_array_almost_equal(A2, A[-1, -3:-1][::-1])
        np.testing.assert_array_almost_equal(A1, A2)

    def test_r2_array_last_site_backward_max_distance(self):
        ts, _ = self.ts()
        ldc = tskit.LdCalculator(ts)
        a = ts.num_sites - 1
        A1 = ldc.r2_array(a, direction=-1, max_distance=3)
        A2 = ts_r2_array(ts, a, direction=-1, max_distance=3)
        np.testing.assert_array_almost_equal(A1, A2)

    @pytest.mark.parametrize("max_sites", [0, 1, 2])
    def test_r2_array_forward_max_sites_zero(self, max_sites):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        site = ts.num_sites // 2
        A1 = ldc.r2_array(site, direction=1, max_sites=max_sites)
        assert A1.shape[0] == max_sites
        A2 = ts_r2_array(ts, site, direction=1, max_sites=max_sites)
        assert A2.shape[0] == max_sites

    @pytest.mark.parametrize("max_sites", [0, 1, 2])
    def test_r2_array_backward_max_sites_zero(self, max_sites):
        ts, A = self.ts()
        ldc = tskit.LdCalculator(ts)
        site = ts.num_sites // 2
        A1 = ldc.r2_array(site, direction=-1, max_sites=max_sites)
        assert A1.shape[0] == max_sites
        A2 = ts_r2_array(ts, site, direction=-1, max_sites=max_sites)
        assert A2.shape[0] == max_sites


class TestLdOneSitePerTree(BaseTestLd):
    @tests.cached_example
    def ts(self):
        ts = msprime.sim_ancestry(
            5, sequence_length=10, recombination_rate=0.1, random_seed=1234
        )
        assert ts.num_trees > 3

        tables = ts.dump_tables()
        for tree in ts.trees():
            site = tables.sites.add_row(tree.interval[0], ancestral_state="A")
            # Put the mutation somewhere deep in the tree
            node = tree.preorder()[2]
            tables.mutations.add_row(site=site, node=node, derived_state="B")
        ts = tables.tree_sequence()
        # Return the full f2 matrix also
        return ts, get_r2_matrix(ts)


class TestLdAllSitesOneTree(BaseTestLd):
    @tests.cached_example
    def ts(self):
        ts = msprime.sim_ancestry(
            5, sequence_length=10, recombination_rate=0.1, random_seed=1234
        )
        assert ts.num_trees > 3

        tables = ts.dump_tables()
        tree = ts.at(5)
        pos = np.linspace(tree.interval[0], tree.interval[1], num=10, endpoint=False)
        for x, node in zip(pos, tree.preorder()[1:]):
            site = tables.sites.add_row(x, ancestral_state="A")
            tables.mutations.add_row(site=site, node=node, derived_state="B")
        ts = tables.tree_sequence()
        return ts, get_r2_matrix(ts)


class TestLdSitesEveryOtherTree(BaseTestLd):
    @tests.cached_example
    def ts(self):
        ts = msprime.sim_ancestry(
            5, sequence_length=20, recombination_rate=0.1, random_seed=1234
        )
        assert ts.num_trees > 5

        tables = ts.dump_tables()
        for tree in ts.trees():
            if tree.index % 2 == 0:
                pos = np.linspace(*tree.interval, num=2, endpoint=False)
                for x, node in zip(pos, tree.preorder()[1:]):
                    site = tables.sites.add_row(x, ancestral_state="A")
                    tables.mutations.add_row(site=site, node=node, derived_state="B")
        ts = tables.tree_sequence()
        return ts, get_r2_matrix(ts)


class TestLdErrors:
    def test_multi_mutations(self):
        tables = tskit.TableCollection(2)
        tables.nodes.add_row(flags=tskit.NODE_IS_SAMPLE, time=0)
        tables.sites.add_row(position=0, ancestral_state="A")
        tables.sites.add_row(position=1, ancestral_state="A")
        tables.mutations.add_row(site=0, node=0, derived_state="C")
        tables.mutations.add_row(site=0, node=0, derived_state="T", parent=0)
        tables.mutations.add_row(site=1, node=0, derived_state="C")
        ts = tables.tree_sequence()
        ldc = tskit.LdCalculator(ts)
        with pytest.raises(tskit.LibraryError, match="Only infinite sites mutations"):
            ldc.r2(0, 1)
        with pytest.raises(tskit.LibraryError, match="Only infinite sites mutations"):
            ldc.r2(1, 0)

    @pytest.mark.parametrize("state", ["", "A", "AAAA", "💩"])
    def test_silent_mutations(self, state):
        tables = tskit.TableCollection(2)
        tables.nodes.add_row(flags=tskit.NODE_IS_SAMPLE, time=0)
        tables.sites.add_row(position=0, ancestral_state=state)
        tables.sites.add_row(position=1, ancestral_state="A")
        tables.mutations.add_row(site=0, node=0, derived_state=state)
        tables.mutations.add_row(site=1, node=0, derived_state="C")
        ts = tables.tree_sequence()
        ldc = tskit.LdCalculator(ts)
        with pytest.raises(tskit.LibraryError, match="Silent mutations not supported"):
            ldc.r2(0, 1)
        with pytest.raises(tskit.LibraryError, match="Silent mutations not supported"):
            ldc.r2(1, 0)


class TestLdCalculator:
    """
    Tests for the LdCalculator class.
    """

    num_test_sites = 50

    def verify_matrix(self, ts):
        m = ts.get_num_sites()
        ldc = tskit.LdCalculator(ts)
        A = ldc.get_r2_matrix()
        assert A.shape == (m, m)
        B = get_r2_matrix(ts)
        assert np.allclose(A, B)

        # Now look at each row in turn, and verify it's the same
        # when we use get_r2 directly.
        for j in range(m):
            a = ldc.get_r2_array(j, direction=tskit.FORWARD)
            b = A[j, j + 1 :]
            assert a.shape[0] == m - j - 1
            assert b.shape[0] == m - j - 1
            assert np.allclose(a, b)
            a = ldc.get_r2_array(j, direction=tskit.REVERSE)
            b = A[j, :j]
            assert a.shape[0] == j
            assert b.shape[0] == j
            assert np.allclose(a[::-1], b)

        # Now check every cell in the matrix in turn.
        for j in range(m):
            for k in range(m):
                assert ldc.get_r2(j, k) == pytest.approx(A[j, k])

    def verify_max_distance(self, ts):
        """
        Verifies that the max_distance parameter works as expected.
        """
        mutations = list(ts.mutations())
        ldc = tskit.LdCalculator(ts)
        A = ldc.get_r2_matrix()
        j = len(mutations) // 2
        for k in range(j):
            x = (
                ts.site(mutations[j + k].site).position
                - ts.site(mutations[j].site).position
            )
            a = ldc.get_r2_array(j, max_distance=x)
            assert a.shape[0] == k
            assert np.allclose(A[j, j + 1 : j + 1 + k], a)
            x = (
                ts.site(mutations[j].site).position
                - ts.site(mutations[j - k].site).position
            )
            a = ldc.get_r2_array(j, max_distance=x, direction=tskit.REVERSE)
            assert a.shape[0] == k
            assert np.allclose(A[j, j - k : j], a[::-1])
        L = ts.get_sequence_length()
        m = len(mutations)
        a = ldc.get_r2_array(0, max_distance=L)
        assert a.shape[0] == m - 1
        assert np.allclose(A[0, 1:], a)
        a = ldc.get_r2_array(m - 1, max_distance=L, direction=tskit.REVERSE)
        assert a.shape[0] == m - 1
        assert np.allclose(A[m - 1, :-1], a[::-1])

    def verify_max_mutations(self, ts):
        """
        Verifies that the max mutations parameter works as expected.
        """
        mutations = list(ts.mutations())
        ldc = tskit.LdCalculator(ts)
        A = ldc.get_r2_matrix()
        j = len(mutations) // 2
        for k in range(j):
            a = ldc.get_r2_array(j, max_mutations=k)
            assert a.shape[0] == k
            assert np.allclose(A[j, j + 1 : j + 1 + k], a)
            a = ldc.get_r2_array(j, max_mutations=k, direction=tskit.REVERSE)
            assert a.shape[0] == k
            assert np.allclose(A[j, j - k : j], a[::-1])

    def test_single_tree_simulated_mutations(self):
        ts = msprime.simulate(20, mutation_rate=10, random_seed=15)
        ts = tsutil.subsample_sites(ts, self.num_test_sites)
        self.verify_matrix(ts)
        self.verify_max_distance(ts)

    def test_deprecated_get_aliases(self):
        ts = msprime.simulate(20, mutation_rate=10, random_seed=15)
        ts = tsutil.subsample_sites(ts, self.num_test_sites)
        ldc = tskit.LdCalculator(ts)
        A = ldc.get_r2_matrix()
        B = ldc.r2_matrix()
        assert np.array_equal(A, B)
        a = ldc.get_r2_array(0)
        b = ldc.r2_array(0)
        assert np.array_equal(a, b)
        assert ldc.get_r2(0, 1) == ldc.r2(0, 1)

    def test_deprecated_max_mutations_alias(self):
        ts = msprime.simulate(2, mutation_rate=0.1, random_seed=15)
        ldc = tskit.LdCalculator(ts)
        with pytest.raises(ValueError, match="deprecated synonym"):
            ldc.r2_array(0, max_sites=1, max_mutations=1)

    def test_single_tree_regular_mutations(self):
        ts = msprime.simulate(self.num_test_sites, length=self.num_test_sites)
        ts = tsutil.insert_branch_mutations(ts)
        # We don't support back mutations, so this should fail.
        with pytest.raises(_tskit.LibraryError):
            self.verify_matrix(ts)
        with pytest.raises(_tskit.LibraryError):
            self.verify_max_distance(ts)

    def test_tree_sequence_regular_mutations(self):
        ts = msprime.simulate(
            self.num_test_sites, recombination_rate=1, length=self.num_test_sites
        )
        assert ts.get_num_trees() > 10
        t = ts.dump_tables()
        t.sites.reset()
        t.mutations.reset()
        for j in range(self.num_test_sites):
            site_id = len(t.sites)
            t.sites.add_row(position=j, ancestral_state="0")
            t.mutations.add_row(site=site_id, derived_state="1", node=j)
        ts = t.tree_sequence()
        self.verify_matrix(ts)
        self.verify_max_distance(ts)

    def test_tree_sequence_simulated_mutations(self):
        ts = msprime.simulate(20, mutation_rate=10, recombination_rate=10)
        assert ts.get_num_trees() > 10
        ts = tsutil.subsample_sites(ts, self.num_test_sites)
        self.verify_matrix(ts)
        self.verify_max_distance(ts)
        self.verify_max_mutations(ts)


def set_partitions(collection):
    """
    Returns an iterator over all partitions of the specified set.

    From https://stackoverflow.com/questions/19368375/set-partitions-in-python
    """
    if len(collection) == 1:
        yield [collection]
    else:
        first = collection[0]
        for smaller in set_partitions(collection[1:]):
            for n, subset in enumerate(smaller):
                yield smaller[:n] + [[first] + subset] + smaller[n + 1 :]
            yield [[first]] + smaller


def naive_mean_descendants(ts, reference_sets):
    """
    Straightforward implementation of mean sample ancestry by iterating
    over the trees and nodes in each tree.
    """
    # TODO generalise this to allow arbitrary nodes, not just samples.
    C = np.zeros((ts.num_nodes, len(reference_sets)))
    T = np.zeros(ts.num_nodes)
    tree_iters = [ts.trees(tracked_samples=sample_set) for sample_set in reference_sets]
    for _ in range(ts.num_trees):
        trees = [next(tree_iter) for tree_iter in tree_iters]
        span = trees[0].span
        for node in trees[0].nodes():
            num_samples = trees[0].num_samples(node)
            if num_samples > 0:
                for j, tree in enumerate(trees):
                    C[node, j] += span * tree.num_tracked_samples(node)
                T[node] += span
    for node in range(ts.num_nodes):
        if T[node] > 0:
            C[node] /= T[node]
    return C


class TestMeanDescendants:
    """
    Tests the TreeSequence.mean_descendants method.
    """

    def verify(self, ts, reference_sets):
        C1 = naive_mean_descendants(ts, reference_sets)
        C2 = tsutil.mean_descendants(ts, reference_sets)
        C3 = ts.mean_descendants(reference_sets)
        assert C1.shape == C2.shape
        assert np.allclose(C1, C2)
        assert np.allclose(C1, C3)
        return C1

    def test_two_populations_high_migration(self):
        ts = msprime.simulate(
            population_configurations=[
                msprime.PopulationConfiguration(8),
                msprime.PopulationConfiguration(8),
            ],
            migration_matrix=[[0, 1], [1, 0]],
            recombination_rate=3,
            random_seed=5,
        )
        assert ts.num_trees > 1
        self.verify(ts, [ts.samples(0), ts.samples(1)])

    def test_single_tree(self):
        ts = msprime.simulate(6, random_seed=1)
        S = [range(3), range(3, 6)]
        C = self.verify(ts, S)
        for j, samples in enumerate(S):
            tree = next(ts.trees(tracked_samples=samples))
            for u in tree.nodes():
                assert tree.num_tracked_samples(u) == C[u, j]

    def test_single_tree_partial_samples(self):
        ts = msprime.simulate(6, random_seed=1)
        S = [range(3), range(3, 4)]
        C = self.verify(ts, S)
        for j, samples in enumerate(S):
            tree = next(ts.trees(tracked_samples=samples))
            for u in tree.nodes():
                assert tree.num_tracked_samples(u) == C[u, j]

    def test_single_tree_all_sample_sets(self):
        ts = msprime.simulate(6, random_seed=1)
        for S in set_partitions(list(range(ts.num_samples))):
            C = self.verify(ts, S)
            for j, samples in enumerate(S):
                tree = next(ts.trees(tracked_samples=samples))
                for u in tree.nodes():
                    assert tree.num_tracked_samples(u) == C[u, j]

    def test_many_trees_all_sample_sets(self):
        ts = msprime.simulate(6, recombination_rate=2, random_seed=1)
        assert ts.num_trees > 2
        for S in set_partitions(list(range(ts.num_samples))):
            self.verify(ts, S)

    def test_wright_fisher_unsimplified_all_sample_sets(self):
        tables = wf.wf_sim(
            4,
            5,
            seed=1,
            deep_history=False,
            initial_generation_samples=False,
            num_loci=10,
        )
        tables.sort()
        ts = tables.tree_sequence()
        for S in set_partitions(list(ts.samples())):
            self.verify(ts, S)

    def test_wright_fisher_unsimplified(self):
        tables = wf.wf_sim(
            20,
            15,
            seed=1,
            deep_history=False,
            initial_generation_samples=False,
            num_loci=20,
        )
        tables.sort()
        ts = tables.tree_sequence()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])

    def test_wright_fisher_simplified(self):
        tables = wf.wf_sim(
            30,
            10,
            seed=1,
            deep_history=False,
            initial_generation_samples=False,
            num_loci=5,
        )
        tables.sort()
        ts = tables.tree_sequence()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])


def naive_genealogical_nearest_neighbours(ts, focal, reference_sets):
    # Make sure everything is a sample so we can use the tracked_samples option.
    # This is a limitation of the current API.
    tables = ts.dump_tables()
    tables.nodes.set_columns(
        flags=np.ones_like(tables.nodes.flags), time=tables.nodes.time
    )
    ts = tables.tree_sequence()

    A = np.zeros((len(focal), len(reference_sets)))
    L = np.zeros(len(focal))
    reference_set_map = np.zeros(ts.num_nodes, dtype=int) - 1
    for k, ref_set in enumerate(reference_sets):
        for u in ref_set:
            reference_set_map[u] = k
    tree_iters = [
        ts.trees(tracked_samples=reference_nodes) for reference_nodes in reference_sets
    ]
    for _ in range(ts.num_trees):
        trees = list(map(next, tree_iters))
        length = trees[0].interval.right - trees[0].interval.left
        for j, u in enumerate(focal):
            focal_node_set = reference_set_map[u]
            # delta(u) = 1 if u exists in any of the reference sets; 0 otherwise
            delta = int(focal_node_set != -1)
            v = u
            while v != tskit.NULL:
                total = sum(tree.num_tracked_samples(v) for tree in trees)
                if total > delta:
                    break
                v = trees[0].parent(v)
            if v != tskit.NULL:
                for k, tree in enumerate(trees):
                    # If the focal node is in the current set, we subtract its
                    # contribution from the numerator
                    n = tree.num_tracked_samples(v) - (k == focal_node_set)
                    # If the focal node is in *any* reference set, we subtract its
                    # contribution from the demoninator.
                    A[j, k] += length * n / (total - delta)
                L[j] += length
    # Normalise by the accumulated value for each focal node.
    index = L > 0
    L = L[index]
    L = L.reshape((L.shape[0], 1))
    A[index, :] /= L
    return A


def parse_time_windows(ts, time_windows):
    if time_windows is None:
        time_windows = [0.0, ts.max_root_time]
    return np.array(time_windows)


def windowed_genealogical_nearest_neighbours(
    ts,
    focal,
    reference_sets,
    windows=None,
    time_windows=None,
    span_normalise=True,
    time_normalise=True,
):
    """
    genealogical_nearest_neighbours with support for span- and time-based windows
    """
    reference_set_map = np.full(ts.num_nodes, tskit.NULL, dtype=int)
    for k, reference_set in enumerate(reference_sets):
        for u in reference_set:
            if reference_set_map[u] != tskit.NULL:
                raise ValueError("Duplicate value in reference sets")
            reference_set_map[u] = k

    windows_used = windows is not None
    time_windows_used = time_windows is not None
    windows = ts.parse_windows(windows)
    num_windows = windows.shape[0] - 1
    time_windows = parse_time_windows(ts, time_windows)
    num_time_windows = time_windows.shape[0] - 1
    A = np.zeros((num_windows, num_time_windows, len(focal), len(reference_sets)))
    K = len(reference_sets)
    parent = np.full(ts.num_nodes, tskit.NULL, dtype=int)
    sample_count = np.zeros((ts.num_nodes, K), dtype=int)
    time = ts.tables.nodes.time
    norm = np.zeros((num_windows, num_time_windows, len(focal)))

    # Set the initial conditions.
    for j in range(K):
        sample_count[reference_sets[j], j] = 1

    window_index = 0
    for (t_left, t_right), edges_out, edges_in in ts.edge_diffs():
        for edge in edges_out:
            parent[edge.child] = tskit.NULL
            v = edge.parent
            while v != tskit.NULL:
                sample_count[v] -= sample_count[edge.child]
                v = parent[v]
        for edge in edges_in:
            parent[edge.child] = edge.parent
            v = edge.parent
            while v != tskit.NULL:
                sample_count[v] += sample_count[edge.child]
                v = parent[v]

        # Update the windows
        assert window_index < num_windows
        while windows[window_index] < t_right and window_index + 1 <= num_windows:
            w_left = windows[window_index]
            w_right = windows[window_index + 1]
            left = max(t_left, w_left)
            right = min(t_right, w_right)
            span = right - left
            # Process this tree.
            for j, u in enumerate(focal):
                focal_reference_set = reference_set_map[u]
                delta = int(focal_reference_set != tskit.NULL)
                p = u
                while p != tskit.NULL:
                    total = np.sum(sample_count[p])
                    if total > delta:
                        break
                    p = parent[p]
                if p != tskit.NULL:
                    scale = span / (total - delta)
                    time_index = np.searchsorted(time_windows, time[p]) - 1
                    if 0 <= time_index < num_time_windows:
                        for k in range(len(reference_sets)):
                            n = sample_count[p, k] - int(focal_reference_set == k)
                            A[window_index, time_index, j, k] += n * scale
                        norm[window_index, time_index, j] += span
            assert span > 0
            if w_right <= t_right:
                window_index += 1
            else:
                # This interval crosses a tree boundary, so we update it again
                # in the next tree
                break

    # Reshape norm depending on normalization selected
    # Return NaN when normalisation value is 0
    if span_normalise and time_normalise:
        reshaped_norm = norm.reshape((num_windows, num_time_windows, len(focal), 1))
    elif span_normalise and not time_normalise:
        norm = np.sum(norm, axis=1)
        reshaped_norm = norm.reshape((num_windows, 1, len(focal), 1))
    elif time_normalise and not span_normalise:
        norm = np.sum(norm, axis=0)
        reshaped_norm = norm.reshape((1, num_time_windows, len(focal), 1))

    with np.errstate(invalid="ignore", divide="ignore"):
        A /= reshaped_norm
    A[np.all(A == 0, axis=3)] = np.nan

    # Remove dimension for windows and/or time_windows if parameter is None
    if not windows_used and time_windows_used:
        A = A.reshape((num_time_windows, len(focal), len(reference_sets)))
    elif not time_windows_used and windows_used:
        A = A.reshape((num_windows, len(focal), len(reference_sets)))
    elif not windows_used and not time_windows_used:
        A = A.reshape((len(focal), len(reference_sets)))
    return A


class TestGenealogicalNearestNeighbours:
    """
    Tests the TreeSequence.genealogical_nearest_neighbours method.
    """

    #
    #          8
    #         / \
    #        /   \
    #       /     \
    #      7       \
    #     / \       6
    #    /   5     / \
    #   /   / \   /   \
    #  4   0   1 2     3
    small_tree_ex_nodes = """\
    id      is_sample   population      time
    0       1       0               0.00000000000000
    1       1       0               0.00000000000000
    2       1       0               0.00000000000000
    3       1       0               0.00000000000000
    4       1       0               0.00000000000000
    5       0       0               0.14567111023387
    6       0       0               0.21385545626353
    7       0       0               0.43508024345063
    8       0       0               1.60156352971203
    """
    small_tree_ex_edges = """\
    id      left            right           parent  child
    0       0.00000000      1.00000000      5       0,1
    1       0.00000000      1.00000000      6       2,3
    2       0.00000000      1.00000000      7       4,5
    3       0.00000000      1.00000000      8       6,7
    """

    def verify(self, ts, reference_sets, focal=None):
        if focal is None:
            focal = [u for refset in reference_sets for u in refset]
        A1 = naive_genealogical_nearest_neighbours(ts, focal, reference_sets)
        A2 = tsutil.genealogical_nearest_neighbours(ts, focal, reference_sets)
        A3 = ts.genealogical_nearest_neighbours(focal, reference_sets)
        A4 = ts.genealogical_nearest_neighbours(focal, reference_sets, num_threads=3)
        A5 = windowed_genealogical_nearest_neighbours(ts, focal, reference_sets)
        assert np.array_equal(A3, A4)
        assert A1.shape == A2.shape
        assert A1.shape == A3.shape
        assert np.allclose(A1, A2)
        assert np.allclose(A1, A3)
        mask = ~np.isnan(A5)
        assert np.sum(mask) > 0 or ts.num_edges == 0
        assert np.allclose(A1[mask], A5[mask])
        assert np.allclose(A5[mask], A2[mask])
        assert np.allclose(A5[mask], A3[mask])

        if ts.num_edges > 0 and all(ts.node(u).is_sample() for u in focal):
            # When the focal nodes are samples, we can assert some stronger properties.
            assert np.allclose(np.sum(A1, axis=1), 1)
            assert np.allclose(np.sum(A5, axis=1), 1)
        return A1

    def test_simple_example_all_samples(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0])
        assert list(A[0]) == [1, 0]
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [4])
        assert list(A[0]) == [1, 0]
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [2])
        assert list(A[0]) == [0, 1]
        A = self.verify(ts, [[0, 2], [1, 3, 4]], [0])
        assert list(A[0]) == [0, 1]
        A = self.verify(ts, [[0, 2], [1, 3, 4]], [4])
        assert list(A[0]) == [0.5, 0.5]

    def test_simple_example_missing_samples(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        A = self.verify(ts, [[0, 1], [2, 4]], [3])
        assert list(A[0]) == [0, 1]
        A = self.verify(ts, [[0, 1], [2, 4]], [2])
        assert np.allclose(A[0], [2 / 3, 1 / 3])

    def test_simple_example_internal_focal_node(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        focal = [7]  # An internal node
        reference_sets = [[4, 0, 1], [2, 3]]
        GNN = naive_genealogical_nearest_neighbours(ts, focal, reference_sets)
        assert np.allclose(GNN[0], np.array([1.0, 0.0]))
        GNN = tsutil.genealogical_nearest_neighbours(ts, focal, reference_sets)
        assert np.allclose(GNN[0], np.array([1.0, 0.0]))
        GNN = ts.genealogical_nearest_neighbours(focal, reference_sets)
        assert np.allclose(GNN[0], np.array([1.0, 0.0]))
        focal = [8]  # The root
        GNN = naive_genealogical_nearest_neighbours(ts, focal, reference_sets)
        assert np.allclose(GNN[0], np.array([0.6, 0.4]))
        GNN = tsutil.genealogical_nearest_neighbours(ts, focal, reference_sets)
        assert np.allclose(GNN[0], np.array([0.6, 0.4]))
        GNN = ts.genealogical_nearest_neighbours(focal, reference_sets)
        assert np.allclose(GNN[0], np.array([0.6, 0.4]))

    def test_two_populations_high_migration(self):
        ts = msprime.simulate(
            population_configurations=[
                msprime.PopulationConfiguration(18),
                msprime.PopulationConfiguration(18),
            ],
            migration_matrix=[[0, 1], [1, 0]],
            recombination_rate=8,
            random_seed=5,
        )
        assert ts.num_trees > 1
        self.verify(ts, [ts.samples(0), ts.samples(1)])

    def test_single_tree(self):
        ts = msprime.simulate(6, random_seed=1)
        S = [range(3), range(3, 6)]
        self.verify(ts, S)

    def test_single_tree_internal_reference_sets(self):
        ts = msprime.simulate(10, random_seed=1)
        tree = ts.first()
        S = [[u] for u in tree.children(tree.root)]
        self.verify(ts, S, ts.samples())

    def test_single_tree_all_nodes(self):
        ts = msprime.simulate(10, random_seed=1)
        S = [np.arange(ts.num_nodes, dtype=np.int32)]
        self.verify(ts, S, np.arange(ts.num_nodes, dtype=np.int32))

    def test_single_tree_partial_samples(self):
        ts = msprime.simulate(6, random_seed=1)
        S = [range(3), range(3, 4)]
        self.verify(ts, S)

    def test_single_tree_all_sample_sets(self):
        ts = msprime.simulate(6, random_seed=1)
        for S in set_partitions(list(range(ts.num_samples))):
            self.verify(ts, S)

    def test_many_trees_all_sample_sets(self):
        ts = msprime.simulate(6, recombination_rate=2, random_seed=1)
        assert ts.num_trees > 2
        for S in set_partitions(list(range(ts.num_samples))):
            self.verify(ts, S)

    def test_many_trees_sequence_length(self):
        for L in [0.5, 1.5, 3.3333]:
            ts = msprime.simulate(6, length=L, recombination_rate=2, random_seed=1)
            self.verify(ts, [range(3), range(3, 6)])

    def test_many_trees_all_nodes(self):
        ts = msprime.simulate(6, length=4, recombination_rate=2, random_seed=1)
        S = [np.arange(ts.num_nodes, dtype=np.int32)]
        self.verify(ts, S, np.arange(ts.num_nodes, dtype=np.int32))

    def test_wright_fisher_unsimplified_all_sample_sets(self):
        tables = wf.wf_sim(
            4,
            5,
            seed=1,
            deep_history=True,
            initial_generation_samples=False,
            num_loci=10,
        )
        tables.sort()
        ts = tables.tree_sequence()
        for S in set_partitions(list(ts.samples())):
            self.verify(ts, S)

    def test_wright_fisher_unsimplified(self):
        tables = wf.wf_sim(
            20,
            15,
            seed=1,
            deep_history=True,
            initial_generation_samples=False,
            num_loci=20,
        )
        tables.sort()
        ts = tables.tree_sequence()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])

    def test_wright_fisher_initial_generation(self):
        tables = wf.wf_sim(
            20,
            15,
            seed=1,
            deep_history=True,
            initial_generation_samples=True,
            num_loci=20,
        )
        tables.sort()
        tables.simplify()
        ts = tables.tree_sequence()
        samples = ts.samples()
        founders = [u for u in samples if ts.node(u).time > 0]
        samples = [u for u in samples if ts.node(u).time == 0]
        self.verify(ts, [founders[:10], founders[10:]], samples)

    def test_wright_fisher_initial_generation_no_deep_history(self):
        tables = wf.wf_sim(
            20,
            15,
            seed=2,
            deep_history=False,
            initial_generation_samples=True,
            num_loci=20,
        )
        tables.sort()
        tables.simplify()
        ts = tables.tree_sequence()
        samples = ts.samples()
        founders = [u for u in samples if ts.node(u).time > 0]
        samples = [u for u in samples if ts.node(u).time == 0]
        self.verify(ts, [founders[:10], founders[10:]], samples)

    def test_wright_fisher_unsimplified_multiple_roots(self):
        tables = wf.wf_sim(
            20,
            15,
            seed=1,
            deep_history=False,
            initial_generation_samples=False,
            num_loci=20,
        )
        tables.sort()
        ts = tables.tree_sequence()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])

    def test_wright_fisher_simplified(self):
        tables = wf.wf_sim(
            31,
            10,
            seed=1,
            deep_history=True,
            initial_generation_samples=False,
            num_loci=5,
        )
        tables.sort()
        ts = tables.tree_sequence().simplify()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])

    def test_wright_fisher_simplified_multiple_roots(self):
        tables = wf.wf_sim(
            31,
            10,
            seed=1,
            deep_history=False,
            initial_generation_samples=False,
            num_loci=5,
        )
        tables.sort()
        ts = tables.tree_sequence()
        samples = ts.samples()
        self.verify(ts, [samples[:10], samples[10:]])

    def test_empty_ts(self):
        tables = tskit.TableCollection(1.0)
        tables.nodes.add_row(1, 0)
        tables.nodes.add_row(1, 0)
        ts = tables.tree_sequence()
        self.verify(ts, [[0], [1]])


class TestWindowedGenealogicalNearestNeighbours(TestGenealogicalNearestNeighbours):
    """
    Tests the TreeSequence.genealogical_nearest_neighbours method.
    """

    #               .    5
    #               .   / \
    #        4      .  |   4
    #       / \     .  |   |\
    #      3   \    .  |   | \
    #     / \   \   .  |   |  \
    #   [0] [1] [2] . [0] [1] [2]
    #
    two_tree_nodes = """\
    id      is_sample   time
    0       1           0
    1       1           0
    2       1           0
    3       0           1
    4       0           2
    5       0           3
    """
    two_tree_edges = """\
    left    right   parent  child
    0       0.2     3       0,1
    0       1       4       2
    0       0.2     4       3
    0.2     1       4       1
    0.2     1       5       0,4
    """

    def get_two_tree_ts(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.two_tree_nodes),
            edges=io.StringIO(self.two_tree_edges),
            strict=False,
        )
        return ts

    def verify(self, ts, reference_sets, focal=None, windows=None, time_windows=None):
        if focal is None:
            focal = [u for refset in reference_sets for u in refset]
        gnn = windowed_genealogical_nearest_neighbours(
            ts, focal, reference_sets, windows, time_windows
        )
        if windows is not None:
            windows_len = len(windows) - 1
        if time_windows is not None:
            time_windows_len = len(time_windows) - 1
        if windows is None and time_windows is None:
            assert np.array_equal(gnn.shape, [len(focal), len(reference_sets)])
        elif windows is None and time_windows is not None:
            assert np.array_equal(
                gnn.shape, [time_windows_len, len(focal), len(reference_sets)]
            )
        elif windows is not None and time_windows is None:
            assert np.array_equal(
                gnn.shape, [windows_len, len(focal), len(reference_sets)]
            )
        else:
            assert np.array_equal(
                gnn.shape,
                [windows_len, time_windows_len, len(focal), len(reference_sets)],
            )

        return gnn

    def test_one_tree_windows(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 1])
        assert np.allclose(A, [[[[1, 0]]]])
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 0.5, 1])
        assert np.allclose(A, [[[[1.0, 0.0]]], [[[1.0, 0.0]]]])
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 0.5, 0.6, 1])
        assert np.allclose(A, [[[[1.0, 0.0]]], [[[1.0, 0.0]]], [[[1.0, 0.0]]]])

    def test_two_tree_windows(self):
        ts = self.get_two_tree_ts()
        A = self.verify(ts, [[0, 1], [2]], [0], [0, 1])
        assert np.allclose(A, [[[0.6, 0.4]]])
        A = self.verify(ts, [[0, 1], [2]], [0], [0, 0.2, 1])
        assert np.allclose(A, [[[1.0, 0.0]], [[0.5, 0.5]]])
        A = self.verify(ts, [[0, 1], [2]], [0], [0, 0.2, 0.5, 1])
        assert np.allclose(A, [[[1.0, 0.0]], [[0.5, 0.5]], [[0.5, 0.5]]])

    def test_one_tree_time_windows(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], None, [0, ts.max_root_time])
        assert np.allclose(A, [[[1, 0]]])
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], None, [1, 2])
        assert np.allclose(A, [[[np.nan, np.nan]]], equal_nan=True)
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], None, [0, 0.1])
        assert np.allclose(A, [[[np.nan, np.nan]]], equal_nan=True)

    def test_two_tree_time_windows(self):
        ts = self.get_two_tree_ts()
        A = self.verify(ts, [[0, 1], [2]], [0], None, [0, ts.max_root_time])
        assert np.allclose(A, [[[0.6, 0.4]]])
        A = self.verify(ts, [[0, 1], [2]], [0], None, [0, 1.1, ts.max_root_time])
        assert np.allclose(A, [[[1.0, 0.0]], [[0.5, 0.5]]])
        A = self.verify(ts, [[0, 1], [2]], [0], None, [0, 0.5, 1])
        assert np.allclose(A, [[[np.nan, np.nan]], [[1.0, 0.0]]], equal_nan=True)
        A = self.verify(ts, [[0, 1], [2]], [0], None, [1, ts.max_root_time, 10])
        assert np.allclose(A, [[[0.5, 0.5]], [[np.nan, np.nan]]], equal_nan=True)

    def test_one_tree_windows_time_windows(self):
        ts = tskit.load_text(
            nodes=io.StringIO(self.small_tree_ex_nodes),
            edges=io.StringIO(self.small_tree_ex_edges),
            strict=False,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 1], [0, ts.max_root_time])
        assert np.allclose(A, [[[[1, 0]]]])
        A = self.verify(
            ts, [[0, 1], [2, 3, 4]], [0], [0, 0.2, 1], [0, 1.1, ts.max_root_time]
        )
        assert np.allclose(
            A,
            [
                [[[1.0, 0.0]], [[np.nan, np.nan]]],
                [[[1.0, 0.0]], [[np.nan, np.nan]]],
            ],
            equal_nan=True,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 0.2], [0, 0.5, 1])
        assert np.allclose(A, [[[[1.0, 0.0]], [[np.nan, np.nan]]]], equal_nan=True)
        A = self.verify(
            ts, [[0, 1], [2, 3, 4]], [0], [0, 0.2, 1, 1.5], [0, ts.max_root_time, 10]
        )
        assert np.allclose(
            A,
            [
                [[[1.0, 0.0]], [[np.nan, np.nan]]],
                [[[1.0, 0.0]], [[np.nan, np.nan]]],
                [[[np.nan, np.nan]], [[np.nan, np.nan]]],
            ],
            equal_nan=True,
        )

    def test_two_tree_windows_time_windows(self):
        ts = self.get_two_tree_ts()
        A = self.verify(ts, [[0, 1], [2]], [0], [0, 1], [0, ts.max_root_time])
        assert np.allclose(A, [[[[0.6, 0.4]]]])
        A = self.verify(ts, [[0, 1], [2]], [0], [0, 0.2, 1], [0, 1.1, ts.max_root_time])
        assert np.allclose(
            A,
            [
                [[[1.0, 0.0]], [[np.nan, np.nan]]],
                [[[np.nan, np.nan]], [[0.5, 0.5]]],
            ],
            equal_nan=True,
        )
        A = self.verify(ts, [[0, 1], [2, 3, 4]], [0], [0, 0.2, 1], [0, 0.5, 1])
        assert np.allclose(
            A,
            [
                [[[np.nan, np.nan]], [[0.5, 0.5]]],
                [[[np.nan, np.nan]], [[np.nan, np.nan]]],
            ],
            equal_nan=True,
        )

    def test_span_normalise(self):
        ts = self.get_two_tree_ts()
        sample_sets = [[0, 1], [2]]
        focal = [0]
        np.random.seed(5)
        windows = ts.sequence_length * np.array([0.2, 0.4, 0.6, 0.8, 1])
        windows.sort()
        windows[0] = 0.0
        windows[-1] = ts.sequence_length

        result1 = windowed_genealogical_nearest_neighbours(
            ts, focal, sample_sets, windows
        )
        result2 = windowed_genealogical_nearest_neighbours(
            ts, focal, sample_sets, windows, span_normalise=True
        )
        result3 = windowed_genealogical_nearest_neighbours(
            ts, focal, sample_sets, windows, span_normalise=False
        )
        denom = np.diff(windows)[:, np.newaxis, np.newaxis]

        # Test the dimensions are correct
        assert np.array_equal(result1.shape, result2.shape)
        assert np.array_equal(result2.shape, result3.shape)

        # Test normalisation is correct
        assert np.allclose(result1, result2)
        assert np.allclose(result1, result3 / denom)

        # If span_normalised, then sum over all reference sets should be 1
        assert np.allclose(np.sum(result1, axis=2), 1)
        assert np.allclose(np.sum(result2, axis=2), 1)
        # If not span_normalised, then sum over all value is 1
        assert np.allclose(result3.sum(), 1)

    def test_time_normalise(self):
        """
        Testing time_normalise is trickier than span_normalise, as the norm
        depends on the span of the nearest neighbours found in each time grid.
        In this small example, we check which grid nodes 3 and 5 fall in, and use their
        spans to check the time_normalisation.
        """
        ts = self.get_two_tree_ts()
        sample_sets = [[0, 1], [2]]
        focal = [0]
        oldest_node = ts.max_root_time
        time_windows = oldest_node * np.array([0.2, 0.4, 0.6, 0.8, 1])
        time_windows.sort()
        time_windows[0] = 0.0
        time_windows[-1] = oldest_node

        # Determine output_dim of the function
        result1 = windowed_genealogical_nearest_neighbours(
            ts, focal, sample_sets, windows=None, time_windows=time_windows
        )
        result2 = windowed_genealogical_nearest_neighbours(
            ts,
            focal,
            sample_sets,
            windows=None,
            time_windows=time_windows,
            time_normalise=True,
        )
        result3 = windowed_genealogical_nearest_neighbours(
            ts,
            focal,
            sample_sets,
            windows=None,
            time_windows=time_windows,
            time_normalise=False,
        )
        denom = np.zeros(len(time_windows) - 1)
        time_index_3 = np.searchsorted(time_windows, ts.tables.nodes.time[3]) - 1
        time_index_5 = np.searchsorted(time_windows, ts.tables.nodes.time[5]) - 1
        denom[time_index_3] += 0.2
        denom[time_index_5] += 0.8

        # Avoid division by zero
        denom[denom == 0] = 1
        denom = denom[:, np.newaxis, np.newaxis]

        # Test the dimensions are correct
        assert np.array_equal(result1.shape, result2.shape)
        assert np.array_equal(result2.shape, result3.shape)

        # Test normalisation is correct
        assert np.allclose(result1, result2, equal_nan=True)
        assert np.allclose(result1, result3 / denom, equal_nan=True)

        # If time_normalised, then sum over all reference sets should be 1
        # Mask out time intervals that sum to 0
        result1_dim_sum = np.sum(result1, axis=2)
        mask = ~(np.isnan(result1_dim_sum))
        assert np.allclose(
            result1_dim_sum[mask],
            np.ones((len(result1_dim_sum), len(focal)))[mask],
            equal_nan=True,
        )
        result2_dim_sum = np.sum(result2, axis=2)
        mask = ~(np.isnan(result2_dim_sum))
        assert np.allclose(
            result2_dim_sum[mask],
            np.ones((len(result2_dim_sum), len(focal)))[mask],
            equal_nan=True,
        )
        # If not span_normalised, then sum over all value is 1
        assert np.allclose(np.nansum(result3), 1)


def exact_genealogical_nearest_neighbours(ts, focal, reference_sets):
    # Same as above, except we return the per-tree value for a single node.

    # Make sure everyhing is a sample so we can use the tracked_samples option.
    # This is a limitation of the current API.
    tables = ts.dump_tables()
    tables.nodes.set_columns(
        flags=np.ones_like(tables.nodes.flags), time=tables.nodes.time
    )
    ts = tables.tree_sequence()

    A = np.zeros((len(reference_sets), ts.num_trees))
    L = np.zeros(ts.num_trees)
    reference_set_map = np.zeros(ts.num_nodes, dtype=int) - 1
    for k, ref_set in enumerate(reference_sets):
        for u in ref_set:
            reference_set_map[u] = k
    tree_iters = [
        ts.trees(tracked_samples=reference_nodes) for reference_nodes in reference_sets
    ]
    u = focal
    focal_node_set = reference_set_map[u]
    # delta(u) = 1 if u exists in any of the reference sets; 0 otherwise
    delta = int(focal_node_set != -1)
    for _ in range(ts.num_trees):
        trees = list(map(next, tree_iters))
        v = trees[0].parent(u)
        while v != tskit.NULL:
            total = sum(tree.num_tracked_samples(v) for tree in trees)
            if total > delta:
                break
            v = trees[0].parent(v)
        if v != tskit.NULL:
            # The length is only reported where the statistic is defined.
            L[trees[0].index] = trees[0].interval.right - trees[0].interval.left
            for k, tree in enumerate(trees):
                # If the focal node is in the current set, we subtract its
                # contribution from the numerator
                n = tree.num_tracked_samples(v) - (k == focal_node_set)
                # If the focal node is in *any* reference set, we subtract its
                # contribution from the demoninator.
                A[k, tree.index] = n / (total - delta)
    return A, L


def local_gnn(ts, focal, reference_sets):
    # Temporary implementation of the treewise GNN.
    reference_set_map = np.zeros(ts.num_nodes, dtype=int) - 1
    for k, reference_set in enumerate(reference_sets):
        for u in reference_set:
            if reference_set_map[u] != -1:
                raise ValueError("Duplicate value in reference sets")
            reference_set_map[u] = k

    K = len(reference_sets)
    A = np.zeros((len(focal), ts.num_trees, K))
    lefts = np.zeros(ts.num_trees, dtype=float)
    rights = np.zeros(ts.num_trees, dtype=float)
    parent = np.zeros(ts.num_nodes, dtype=int) - 1
    sample_count = np.zeros((ts.num_nodes, K), dtype=int)

    # Set the intitial conditions.
    for j in range(K):
        sample_count[reference_sets[j], j] = 1

    for t, ((left, right), edges_out, edges_in) in enumerate(ts.edge_diffs()):
        for edge in edges_out:
            parent[edge.child] = -1
            v = edge.parent
            while v != -1:
                sample_count[v] -= sample_count[edge.child]
                v = parent[v]
        for edge in edges_in:
            parent[edge.child] = edge.parent
            v = edge.parent
            while v != -1:
                sample_count[v] += sample_count[edge.child]
                v = parent[v]

        # Process this tree.
        for j, u in enumerate(focal):
            focal_reference_set = reference_set_map[u]
            delta = int(focal_reference_set != -1)
            p = parent[u]
            lefts[t] = left
            rights[t] = right
            while p != tskit.NULL:
                total = np.sum(sample_count[p])
                if total > delta:
                    break
                p = parent[p]
            if p != tskit.NULL:
                scale = 1 / (total - delta)
                for k, _reference_set in enumerate(reference_sets):
                    n = sample_count[p, k] - int(focal_reference_set == k)
                    A[j, t, k] = n * scale
    return (A, lefts, rights)


class TestExactGenealogicalNearestNeighbours(TestGenealogicalNearestNeighbours):
    # This is a work in progress - these tests will be adapted to use the
    # treewise GNN when it's implemented.

    def verify(self, ts, reference_sets, focal=None):
        if focal is None:
            focal = [u for refset in reference_sets for u in refset]
        A = ts.genealogical_nearest_neighbours(focal, reference_sets)

        G, lefts, rights = local_gnn(ts, focal, reference_sets)
        for tree in ts.trees():
            assert lefts[tree.index] == tree.interval.left
            assert rights[tree.index] == tree.interval.right

        for j, u in enumerate(focal):
            T, L = exact_genealogical_nearest_neighbours(ts, u, reference_sets)
            assert np.allclose(G[j], T.T)
            # Ignore the cases where the node has no GNNs
            if np.sum(L) > 0:
                mean = np.sum(T * L, axis=1) / np.sum(L)
                assert np.allclose(mean, A[j])
        return A
